/*=============================================================================

				dihed.c

			calculate dihedral angles

Purpose: calculate the angles psi and theta in a protein molecule. 

History:
(1)	17.03.1998	created
(2)	29.03.1998	split into separate files (one function - one file)
(3)	11.10.1998	bond length checking added

========includes:============================================================*/

#include <stdio.h>
#include <stdlib.h>
#include "common.h"
#include "protein.h"
#include "pdb.h"

/*======defines:=============================================================*/

/* See header files, especially common.h ! */

/*======global scope data:===================================================*/

/* No global scope data here, but see header files for some defines */

/*======function prototypes:=================================================*/

void		DihedPrintUsage_ (void);
int		ExtractFlags_ (FlagS *, int, int, char **);
int		PrepareFileNames_ (char *, int, char *, int,
				   FlagS *, int, int, char **);
int		CheckFiles_ (char *, char *);
AtomDataS	*ReadPDBFile_ (long *, AtomFormatS *, char *);
ResidueS	*ExtractResidueData_ (long *, AtomDataS *, long);
DihedS		*CalcDihedralAngles_ (long *, ResidueS *, long, AtomDataS *);
int		DihedWriteResults_ (char *, DihedS *, ResidueS *, long,
				    FlagS *);
void		ErrorMessage_ (char *, char *, char *,
			       char *, char *, char *, char *);

/*======main:================================================================*/

int main (int argc, char *argv[])
{
int			MainF = 0;
FlagS			FSA[3] = {{"-r", 0, 0}, {"-s", 0, 0}, {"-o", 0, 0}};
			/* See usage info for details about flags! */
int			FlagsN = 3, FlagsFoundN;
char			InFileName[STRINGSIZE], OutFileName[STRINGSIZE];
int			n1, n2;
AtomDataS		*ASP = NULL;
long			AtomsN = 0L;
ResidueS		*RSP = NULL;
long			ResiduesN = 0L;
DihedS			*DSP = NULL;
long			DihedsN = 0L;
			/* If changing the format  specification, be sure */
			/* to check the order of data and string lengths: */
AtomFormatS		AtomFmtS = {"ATOM",
				     6, 10, 12, 15, 16, 16, 17, 19, 21, 21,
				    22, 25, 26, 26, 30, 37, 38, 45, 46, 53,
				    54, 59, 60, 65, 72, 75, 76, 77, 78, 79};

/* Print usage info if number of command line arguments is bad: */
if (argc < 3)
	{
	DihedPrintUsage_ ();
	exit (0);
	}

/* Extract flags: */
FlagsFoundN = ExtractFlags_ (FSA, FlagsN, argc, argv);

/* Prepare the file names: */
n1 = sizeof (InFileName);
n2 = sizeof (OutFileName);
if (PrepareFileNames_ (InFileName, n1, OutFileName, n2,
		       FSA, FlagsN, argc, argv) < 0) exit (0);

/* Check files: */
if (CheckFiles_ (InFileName, OutFileName) < 0) exit (0);

/* From this point, large amounts of memory are allocated: */
do
	{
	/* Allocate the memory for atomic data; read pdb file: */
	if (!(ASP = ReadPDBFile_ (&AtomsN, &AtomFmtS, InFileName))) break;
	if (AtomsN <= 0L)
		{
		ErrorMessage_ ("dihed", "main", InFileName,
				"No atoms read!",
				" Check input file!\n", "", "");
		break;
		}
	else printf ("Number of atoms: %ld\n", AtomsN);

	/* Allocate the memory for residue data; extract data: */
	if (!(RSP = ExtractResidueData_ (&ResiduesN, ASP, AtomsN)))
		break;
	if (ResiduesN <= 0L)
		{
		ErrorMessage_ ("dihed", "main", InFileName,
				"No residues recognized!",
				" Check input file!\n", "", "");
		break;
		}
	else printf ("Number of residues: %ld\n", ResiduesN);

	/* Allocate memory for dihedral angles; calculate angles: */
	if (!(DSP = CalcDihedralAngles_ (&DihedsN, RSP, ResiduesN, ASP)))
		break;
	if (DihedsN <= 0L)
		{
		ErrorMessage_ ("dihed", "main", InFileName,
				"Failed to calculate dihedral angles!",
				" (coordinates bad?)\n", "", "");
		break;
		}
	else printf ("Number of dihedral angles: %ld\n", DihedsN);

	/* Write  dihedral  angles, residue  names, */
	/*  residue numbers and chain ID's to file: */
	if (DihedWriteResults_ (OutFileName, DSP, RSP, ResiduesN, FSA) < 0)
		break;

	/* Set the main flag to 1 if everything worked fine: */
	MainF = 1;
	}
while (0);

/* Free memory: */
if (ASP) free (ASP);
if (RSP) free (RSP);
if (DSP) free (DSP);

/* Check the main flag (0 if something went wrong, 1 if everything worked): */
if (!MainF) exit (0);

/* The end: */
return 1;
}

/*===========================================================================*/


