/*=============================================================================

				exseq.c

		extract protein sequence from PDB file

Purpose: Extract sequence from PDB file, using coordinates (not the sequence
	 information from PDB header).

Revision history:
(1)		11.11.1998	created

==========includes:==========================================================*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>

/*======defines:=============================================================*/

#define STRINGSIZE 1000

/*======function prototypes:=================================================*/

void		PrintUsage_ (void);
void		PrintHelp_ (void);
int		CheckFormatFlag_ (int, char **);
int		ExtractFileNames_ (char *, int, char *, int, int, char **);
long		ExtractSequence_ (char *, char *, int);
char		*ExtractSubstring_ (char *, char *, int, int);
int		ConvertAminoAcidName_ (char *, char *);
void		WriteData_ (FILE *, int, int,
			    char *, char *, char *, char *, int);

/*======main:================================================================*/

int main (int argc, char *argv[])
{
int		n;
int		OutputFormat;
char		InFileName[STRINGSIZE], OutFileName[STRINGSIZE];

/* Check is help requested: */
for (n = 1; n < argc; n++)
	{
	if (strstr (argv[n], "-h") == argv[n])
		{
		PrintHelp_ ();
		exit (0);
		}
	}

/* At least three command line arguments expected: */
if (argc < 3)
	{ fprintf (stderr, "\nThree command line arguments expected!\n");
	PrintUsage_ ();
	exit (-1);
	}

/* Check which output format is requested: */
if ((OutputFormat = CheckFormatFlag_ (argc, argv)) < 0)
	{
	fprintf (stderr, "\nUnknown flag(s) found, exiting ...\n");
	fprintf (stderr, "Check flag(s)!\n");
	exit (-2);
	}

/* Extract file names: */
n = STRINGSIZE - 1;
if (ExtractFileNames_ (OutFileName, n, InFileName, n, argc, argv) < 2)
	{
	fprintf (stderr, "\nFailed to extract file names;");
	fprintf (stderr, " check command line arguments!\n");
	PrintUsage_ ();
	exit (-3);
	}

/* Read input pdb file, extract sequence, write to output file: */
if (ExtractSequence_ (OutFileName, InFileName, OutputFormat) < 0) exit (-4);

/* The end: */
return 1;

}

/*======usage info:==========================================================*/

void PrintUsage_ (void)
{
fprintf (stdout, "\n======================================\n");
fprintf (stdout, "Extract protein sequence from pdb file\n");
fprintf (stdout, "======================================\n\n");
fprintf (stdout, "Usage:    exseq [-fotv] infile outfile\n");
fprintf (stdout, "          exseq -h      ... print help\n\n");
}

/*======detailed information:================================================*/

void PrintHelp_ (void)
{
fprintf (stdout,
   "================================================================\n");
fprintf (stdout,
   "The purpose of this program is  to extract  the protein sequence\n");
fprintf (stdout,
   "from PDB file. Four output formats are  available (see options).\n");
fprintf (stdout,
   "----------------------------------------------------------------\n");
fprintf (stdout, "Usage:         exseq [-fotv] infile outfile\n");
fprintf (stdout, "               exseq -h      ... print help\n");
fprintf (stdout, "Options:\n");
fprintf (stdout,
   "               -f        FASTA  (default),  70  codes  per line.\n\n");
fprintf (stdout,
   "               -o        One  letter  code,  40 codes  per line.\n");
fprintf (stdout,
   "                         Subunits  will be  separated in blocks.\n\n");
fprintf (stdout,
   "               -t        Three letter codes, 20 codes  per line.\n");
fprintf (stdout,
   "                         Subunits  will be  separated in blocks.\n\n");
fprintf (stdout,
   "               -v        Verbose output: write global amino acid\n");
fprintf (stdout,
   "                         index, local  amino  acid  index, three\n");
fprintf (stdout,
   "                         letters code,  one letter  code,  amino\n");
fprintf (stdout,
   "                         acid serial number and chain ID.\n\n");
fprintf (stdout,
   "If no option is specified, file will be written in FASTA format.\n");
fprintf (stdout,
   "================================================================\n");
}

/*======check format flag:===================================================*/
/* 0 = one letter code, 40 characters per line (flag -o) */
/* 1 = three letter codes, 20  codes  per line (flag -t) */
/* 2 = three letter codes, one code per  line, (flag -v) */
/*     amino acid  serial number and chain  ID  (letter) */
/*     given for each amino acid code                    */
/* Default: 0 (flag(s) not used at all, or flag -o used) */
int CheckFormatFlag_ (int argc, char *argv[])
{
int		OutputFormat = -1;
int		FlagsN = 0;
int		i;

/* Number of flags: */
for (i = 1; i < argc; i++)
	{
	if (strstr (argv[i], "-") == argv[i]) FlagsN++;
	}

/* Default mode: */
if (FlagsN == 0)
	{
	printf ("FASTA file will be prepared (default option)\n");
	return 0;
	}

/* If there are some flags, recognize requested output format: */
for (i = 1; i < argc; i++)
	{
	if (strcmp (argv[i], "-f") == 0)
		{
		printf ("Flag -f found, FASTA file will be prepared.\n");
		OutputFormat = 0;
		break;
		}
	if (strcmp (argv[i], "-o") == 0)
		{
		printf ("Flag -o found, one letter codes will be prepared.\n");
		OutputFormat = 1;
		break;
		}
	else if (strcmp (argv[i], "-t") == 0)
		{
		printf ("Flag -t found, three letter codes");
		printf (" will be prepared\n");
		OutputFormat = 2;
		break;
		}
	else if (strcmp (argv[i], "-v") == 0)
		{
		printf ("Flag -v found, verbose output will be prepared.\n");
		OutputFormat = 3;
		break;
		}
	}

return OutputFormat;
}

/*======extract file names:==================================================*/

int ExtractFileNames_ (char *OutFileName, int OutFileNameSize,
		       char *InFileName, int InFileNameSize,
		       int argc, char *argv[])
{
int		ReturnValue = 0;
int		i;

/* Scan command line arguments backwards: */
for (i = argc - 1; i > 0; i--)		      /* Do not check program name! */
	{
	if (strstr (argv[i], "-") == argv[i]) continue;
	if (ReturnValue == 0)
		{
		strncpy (OutFileName, argv[i], OutFileNameSize - 1);
		OutFileName[OutFileNameSize - 1] = '\0';
		}
	else if (ReturnValue == 1)
		{
		strncpy (InFileName, argv[i], InFileNameSize - 1);
		InFileName[InFileNameSize - 1] = '\0';
		}
	else break;
	ReturnValue++;
	}

/* Compare file names (should be different): */
if (strcmp (OutFileName, InFileName) == 0)
	{
	fprintf (stderr, "\nInput file name and output file name should be");
	fprintf (stderr, " different!\n");
	return -1;
	}

/* Return the number of file names extracted (two if successful): */
return ReturnValue;
}


/*======extract sequence:====================================================*/

long ExtractSequence_ (char *OutFileName, char *InFileName, int OutputFormat)
{
long		TempFactorsN = 0;
FILE		*InFile, *OutFile;
char		Line[STRINGSIZE];
int		LineLength;
char		AtomID[5] = "ATOM";
char		NewLabel[30], OldLabel[30] = "XXXXXX";
char		NewChainID[30], OldChainID[30] = "*";
char		AAName[30], ShortName[30];
int		LocalN = 0;          /* Number of amino acids in chain */
int		GlobalN = 0;         /* Total number of amino acids */
int		ChainsN = 0;

/* Open input file: */
if ((InFile = fopen (InFileName, "r")) == NULL)
	{
	fprintf (stderr, "\nFailed to open input file!\n");
	return -1;
	}

/* Open output file: */
if ((OutFile = fopen (OutFileName, "w")) == NULL)
	{
        fprintf (stderr, "\nFailed to open output file!\n");
	fclose (InFile);
	return -1;
	}

/* Read line by line: */
LineLength = sizeof (Line);
while (fgets (Line, LineLength, InFile) != NULL)
	{
	/* Skip short lines: */
	if (strlen (Line) < 10) continue;

	/* Check the line (ATOM lines are good ones): */
	if (strstr (Line, AtomID) != Line) continue;

	/* Extract the amino acid label (string with serial number): */
	if (!ExtractSubstring_ (NewLabel, Line, 22, 4)) continue;

	/* Compare this label (serial number) with the old one: */
	if (strcmp (OldLabel, NewLabel) == 0) continue;

	/* Copy label: */
	strncpy (OldLabel, NewLabel, 9);

	/* Extract chain ID, if present: */
	ExtractSubstring_ (NewChainID, Line, 21, 1);

	/* If new chain ID is different from old chain ID, print it to file: */
	if (strcmp (OldChainID, NewChainID) != 0)
		{
		LocalN = 0;
		if (ChainsN != 0)
			{
			fprintf (OutFile, "\n");
			if (OutputFormat != 3) fprintf (OutFile, "\n");
			}
		if (OutputFormat != 0)
			{
			fprintf (OutFile, "#Chain %1s:\n", NewChainID);
			}
		else fprintf (OutFile, ">Chain_%1s:\n", NewChainID);
		ChainsN++;
		}

	/* Copy chain ID: */
	strncpy (OldChainID, NewChainID, 1);

	/* Extract amino acid name: */
	ExtractSubstring_ (AAName, Line, 17, 3);

	/* Check amino acid name; ignore strange amino acid names: */
	if (ConvertAminoAcidName_ (ShortName, AAName) < 0) continue;

	/* Everything worked fine, increase counts: */
	LocalN++;
	GlobalN++;

	/* Write data to output file: */
	WriteData_ (OutFile, GlobalN, LocalN,
		    NewLabel, NewChainID, AAName, ShortName, OutputFormat);
	}

/* One newline at the end: */
fprintf (OutFile, "\n");

/* Close files: */
fclose (InFile);
fclose (OutFile);

/* Return the number of temperature factors: */
return TempFactorsN;
}

/*======extract substring:===================================================*/

char *ExtractSubstring_ (char *dest, char *src, int StartI, int Length)
{
char		*p;

/* Copy substring: */
p = src + StartI;
strncpy (dest, p, Length);
dest[Length] = '\0';

/* Return pointer to dest on success: */
return dest;
}

/*======check amino acid name, convert to one letter code:===================*/
int ConvertAminoAcidName_ (char *dest, char *s)
{
int		i;
static char	GoodName[23][4] = {"ALA", "CYS", "ASP", "GLU", "PHE",
				   "GLY", "HIS", "ILE", "LYS", "LEU",
				   "MET", "ASN", "PRO", "GLN", "ARG",
				   "SER", "THR", "VAL", "TRP", "TYR",
				   "ASX", "GLX", "UNK"};
static char	ShortName[23][2] = {"A",   "C",   "D",   "E",   "F",
				    "G",   "H",   "I",   "K",   "L",
				    "M",   "N",   "P",   "Q",   "R",
				    "S",   "T",   "V",   "W",   "Y",
				    "B",   "Z",   "X"};

/* Convert input string to uppercase, though it should be uppercase already: */
for (i = 0; i < strlen (s); i++) *(s + i) = toupper (*(s + i));

for (i = 0; i < 23; i++)
	{
	if (strcmp (GoodName[i], s) == 0)
		{
		strcpy (dest, ShortName[i]);
		*(dest + 1) = '\0';
		return 1;
		}
	}

/* Return negative integer if amino acid not recognized: */
return -1;
}

/*======write data to output file:===========================================*/

void WriteData_ (FILE *OutFile, int global, int local,	
		 char *serial, char *id, char *name, char *shortname, int type)
{
int		n;

if (type == 0)
	{
	fprintf (OutFile, "%1s", shortname);
	n = local % 70;
	if (n == 0) fprintf (OutFile, "\n");
	}
else if (type == 1)
	{
	fprintf (OutFile, "%1s", shortname);
	n = local % 40;
	if (n == 0) fprintf (OutFile, "\n");
	}
else if (type == 2)
	{
	fprintf (OutFile, "%3s", name);
	n = local % 20;
	if (n == 0) fprintf (OutFile, "\n");
	else fprintf (OutFile, " ");
	}
else if (type == 3)
	{
	fprintf (OutFile, "%6d %5d   %3s %1s  %4s %1s\n",
		 global, local, name, shortname, serial, id);
	}
}

/*===========================================================================*/


