/*=============================================================================

				  rtf.c

			extract temperature factors

Purpose: Extract temperature factors from PDB file. Differences and 
	 second differences ("differences of differences") can be calculated
	 too. This may (hopefully) help to understand the relation between
	 hydrophobicity patterns and secondary structure. 

Revision history:
(1)		22.09.1998	created
(2) 		06.11.1998	significant revision (differences)

==========includes:==========================================================*/

#include <stdio.h>
#include <string.h>

/*======defines:=============================================================*/

#define STRINGSIZE 1000

/*======function prototypes:=================================================*/

void		PrintUsage_ (void);
void		PrintHelp_ (void);
int		Mode_ (int, char **);
int		ExtractFileNames_ (char *, int, char *, int, int, char **);
long		ExtractTempFactors_ (char *, char *, int);
char		*AminoAcidLabel_ (char *, char *);
double		TempFactor_ (char *);

/*======main:================================================================*/

int main (int argc, char *argv[])
{
int		i;
int		ModeFlag;
char		InFileName[1000], OutFileName[1000];

/* Check is help requested: */
for (i = 1; i < argc; i++)
	{
	if (strstr (argv[i], "-h") == argv[i])
		{
		PrintHelp_ ();
		PrintUsage_ ();
		exit (0);
		}
	}

/* At least three command line arguments expected: */
if (argc < 3)
	{
	fprintf (stderr, "\nThree command line arguments expected!\n");
	PrintUsage_ ();
	exit (-1);
	}

/* Check mode: */
if ((ModeFlag = Mode_ (argc, argv)) < 0)
	{
	fprintf (stderr, "\nRequested mode not recognized!");
	fprintf (stderr, " Check flag(s)!\n");
	exit (-2);
	}

/* Extract file names: */
if (ExtractFileNames_ (OutFileName, 999, InFileName, 999, argc, argv) < 2)
	{
	fprintf (stderr, "\nFailed to extract file names;");
	fprintf (stderr, " check command line arguments!\n");
	PrintUsage_ ();
	exit (-3);
	}

/* Read input pdb file, extract temperature factors, write to output file: */
if (ExtractTempFactors_ (OutFileName, InFileName, ModeFlag) < 0) exit (-4);

/* The end: */
return 1;

}

/*======usage info:==========================================================*/

void PrintUsage_ (void)
{
fprintf (stdout, "\n=========================================\n");
fprintf (stdout, "Extract temperature factors from pdb file\n");
fprintf (stdout, "=========================================\n\n");
fprintf (stdout, "Usage:          extf [-ds] infile outfile\n");
fprintf (stdout, "                extf -h    ... print help\n\n");
fprintf (stdout, "Use option -d to print t. f. differences,\n");
fprintf (stdout, "or option -s to print second differences.\n");
}

/*======detailed information:================================================*/

void PrintHelp_ (void)
{
fprintf (stdout,
   "\n=================================================================\n");
fprintf (stdout,
   "The purpose of this program is to extract the temperature factors\n");
fprintf (stdout,
   "from PDB file.  True temperature factors may be replaced  by some\n");
fprintf (stdout,
   "other values (hydrophobicity, for example) using the program rtf.\n");
fprintf (stdout,
   "It is possible to print differences between neighbouring factors,\n");
fprintf (stdout,
   "as  well  as  second  differences  (differences  of differences).\n");
fprintf (stdout,
   "File names should not begin with minus sign [-]!\n");

fprintf (stdout,
   "=================================================================\n");
}

/*======mode requested:======================================================*/

int Mode_ (int argc, char *argv[])
{
int		ReturnValue = -1;
int		FlagsN = 0;
int		i;

/* Number of flags: */
for (i = 1; i < argc; i++)
	{
	if (strstr (argv[i], "-") == argv[i]) FlagsN++;
	}

/* Default mode (print temperature factors, not differences): */
if (FlagsN == 0)
	{
	printf ("Temperature factor values (not differences) will be");
	printf (" extracted!\n");
	return 0;
	}

/* If there are some flags, recognize mode request: */
for (i = 1; i < argc; i++)
	{
	if (strcmp (argv[i], "-d") == 0)
		{
		printf ("Flag -d found, differences will be calculated!\n");
		ReturnValue = 1;
		break;
		}
	else if (strcmp (argv[i], "-s") == 0)
		{
		printf ("Flag -s found, second differences will be");
		printf (" calculated!\n");
		ReturnValue = 2;
		break;
		}
	}

/* Return -1 on failure, positive on success: */
return ReturnValue;
}

/*======extract file names:==================================================*/

int ExtractFileNames_ (char *OutFileName, int OutFileNameSize,
		       char *InFileName, int InFileNameSize,
		       int argc, char *argv[])
{
int		ReturnValue = 0;
int		i;

/* Scan command line arguments backwards: */
for (i = argc - 1; i > 0; i--)		      /* Do not check program name! */
	{
	if (strstr (argv[i], "-") == argv[i]) continue;
	if (ReturnValue == 0)
		{
		strncpy (OutFileName, argv[i], OutFileNameSize - 1);
		OutFileName[OutFileNameSize - 1] = '\0';
		}
	else if (ReturnValue == 1)
		{
		strncpy (InFileName, argv[i], InFileNameSize - 1);
		InFileName[InFileNameSize - 1] = '\0';
		}
	else break;
	ReturnValue++;
	}

/* Compare file names (should be different): */
if (strcmp (OutFileName, InFileName) == 0)
	{
	fprintf (stderr, "\nInput file name and output file name should be");
	fprintf (stderr, " different!\n");
	return -1;
	}

/* Return the number of file names extracted (two if successful): */
return ReturnValue;
}


/*======extract temperature factors:=========================================*/

long ExtractTempFactors_ (char *OutFileName, char *InFileName, int ModeFlag)
{
long		TempFactorsN = 0;
FILE		*InFile, *OutFile;
char		Line[STRINGSIZE];
int		LineLength;
char		AtomID[5] = "ATOM";
char		NewLabel[20], OldLabel[20] = "XXXXXX";
double		NewTF, OldTF = -9999.9, VeryOldTF = -9999.9, Diff;

/* Open input file: */
if ((InFile = fopen (InFileName, "r")) == NULL)
	{
	fprintf (stderr, "\nFailed to open input file!\n");
	return -1;
	}

/* Open output file: */
if ((OutFile = fopen (OutFileName, "w")) == NULL)
	{
        fprintf (stderr, "\nFailed to open output file!\n");
	fclose (InFile);
	return -1;
	}

/* Read line by line: */
LineLength = sizeof (Line);
while (fgets (Line, LineLength, InFile) != NULL)
	{
	/* Skip short lines: */
	if (strlen (Line) < 10) continue;

	/* Check the line (ATOM lines are good ones): */
	if (strstr (Line, AtomID) != Line) continue;

	/* Extract the amino acid label: */
	if (!AminoAcidLabel_ (NewLabel, Line)) continue;

	/* Compare this label with the old one: */
	if (strcmp (OldLabel, NewLabel) == 0) continue;

	/* Copy label: */
	strncpy (OldLabel, NewLabel, 9);

	/* Extract temperature factor: */
	NewTF = TempFactor_ (Line);

	/* Increase count: */
	TempFactorsN++;

	/* Check what should be printed to output file: */
	if (ModeFlag == 0)
		{
		/** Print the order number and temperature factor: **/
		fprintf (OutFile, "%5ld %7.2f\n", TempFactorsN, NewTF);
		}
	else if (ModeFlag == 1)
		{
		/** Calculate  difference,  copy  the **/
		/** temperature factor for later use: **/
		Diff = NewTF - OldTF;
		OldTF = NewTF;
		/** Print the order number minus one and difference: **/
		if (TempFactorsN < 2) continue;
		fprintf (OutFile, "%5ld %7.2f\n", TempFactorsN - 1, Diff);
		}
	else if (ModeFlag == 2)
		{
		/** Calculate the second difference, **/
		/** copy  the  temperature  factors: **/
		Diff = NewTF - 2 * OldTF + VeryOldTF;
		VeryOldTF = OldTF;
		OldTF = NewTF;
		/** Print the order number minus two and difference: **/
		if (TempFactorsN < 3) continue;
		fprintf (OutFile, "%5ld %7.2f\n", TempFactorsN - 2, Diff);
		}
	}

/* Close files: */
fclose (InFile);
fclose (OutFile);

/* Return the number of temperature factors: */
return TempFactorsN;
}

/*======extract amino acid label:============================================*/

char *AminoAcidLabel_ (char *dest, char *src)
{
int		StartI = 21, Length = 5;
char		*p;

/* Check the src string length: */
if (strlen (src) < 26) return NULL;

/* Copy substring: */
p = src + StartI;
strncpy (dest, p, Length);
dest[Length] = '\0';

/* Return value on success: */
return dest;
}

/*======extract temperature factor:==========================================*/

double TempFactor_ (char *Line)
{
double		tf;
char		AuxString[20];
int		StartI = 60, Length = 6;
char		*p;

/* Extract the substring: */
p = Line + StartI;
strncpy (AuxString, p, Length);
AuxString[Length] = '\0';

/* Read double value from substring: */
if (sscanf (AuxString, "%lf", &tf) != 1) return 0.0;

/* Return temperature factor on success: */
return tf;
}

/*===========================================================================*/


