/*=============================================================================

				 neigh.c

	Read STRIDE secondary structure assignments; for the specified
	amino acid, find all neighbours at given  relative  positions. 

-------------------------------------------------------------------------------

Purpose: Read STRIDE output; find all neighbour(s) at specified position(s),
	 for each occurence of selected amino acid.  One, two, three or four
	 neighbour(s) may be specified. Do it for each occurence of selected
	 amino  acid.  Associate  numeric values to neighbours, according to
	 amino acid types. Type neigh -h for more details.

Example 1:

	neigh ARG -3 -4 xyz.ssa xyz.neigh kyte-doo.map

	Find  third and  fourth  neighbour of  arginine, looking in
	reverse  direction of protein sequence (i.e., find residues
	preceding  arginine). The input  file  xyz.ssa  was created
	from  xyz.pdb  using program  STRIDE. The output file  will
	be  xyz.neigh,  and  Kyte-Doolittle  hydrophobicity  values
	(from kyte-doo.map file)  will  be assigned to  neighbours.

Example 2:

	neigh GLY -2 -1 1 2 1occ.ssa 1occ.neigh bulky.map

	Find  four  glicine  neighbours (two before and two after).
	Write side chain  sizes  file  to output  file (side  chain
	sizes are defined in bulky.map file).

Revision history:
(1)		29.11.1998	created

==========includes:==========================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "neigh.h"

/*======defines:=============================================================*/

/*======function prototypes:=================================================*/

void		PrintUsage_ (void);
void		PrintHelp_ (void);
void		ErrorMessage_ (char *, char *, char *,
			       char *, char *, char *, char *);
void		WarningMessage_ (char *, char *, char *,
				 char *, char *, char *, char *);
int		InterpretArgs_ (int *, char *, char *, char *, char *,
				int, char **);
int		IsInteger_ (char *);
int		CheckAminoAcidName_ (AminoAcidNameS *, char *);
int		ReadMapFile_ (double *, AminoAcidNameS *, char *);
int		ExtractDigits_ (char *);
int		RawJob_ (int *, int, double *, AminoAcidNameS *,
			 char *, char *, char *);
long		ReadStructure_ (int *, ResidueDataS **, long *, FILE *);
long		AssignValues_ (ResidueDataS *,
			       long, AminoAcidNameS *, double *);
long		ScanStructure_ (FILE *,
				char *, ResidueDataS *, long, int *, int);
int		LookForward_ (ResidueDataS *, long, long, int);
int		LookBackward_ (ResidueDataS *, long, long, int);

/*======main:================================================================*/

int main (int argc, char *argv[])
{
int		n;
char		SelectedResidue[STRINGSIZE];
char		InFileName[STRINGSIZE], OutFileName[STRINGSIZE],
		MapFileName[STRINGSIZE];
int		size;
AminoAcidNameS	namesSA[23] = {
		{"A", "ALA"}, {"C", "CYS"}, {"L", "LEU"}, {"M", "MET"},
		{"E", "GLU"}, {"Q", "GLN"}, {"H", "HIS"}, {"K", "LYS"},
		{"V", "VAL"}, {"I", "ILE"}, {"F", "PHE"}, {"Y", "TYR"},
		{"W", "TRP"}, {"T", "THR"}, {"G", "GLY"}, {"S", "SER"},
		{"D", "ASP"}, {"N", "ASN"}, {"P", "PRO"}, {"R", "ARG"},
		{"B", "ASX"}, {"Z", "GLX"}, {"X", "UNK"}};
double		valuesA[23] = {
		 1.8,  2.5,  3.8,  1.9,
		-3.5, -3.5, -3.2, -3.9,
		 4.2,  4.5,  2.8, -1.3,
		-0.9, -0.7, -0.4, -0.8,
		-3.5, -3.5, -1.6, -4.5,
		-3.5, -3.5,  0.0};
int		NeighboursN;                         /* Number of neighbours */
int		NeighbourIA[10000];      /* Array to store neighbour indices */

size = STRINGSIZE;

/* Check is help requested: */
for (n = 1; n < argc; n++)
	{
	if (strstr (argv[n], "-h") == argv[n])
		{
		PrintHelp_ ();
		exit (0);
		}
	}

/* At least three command line arguments expected: */
if (argc < 3)
	{
	fprintf (stderr,
		"\nAt least three command line arguments");
	fprintf (stderr,
		" should be specified!\n");
	PrintUsage_ ();
	exit (0);
	}

/* Separate command line arguments: */
/** (Numeric arguments will be stored in NeighbourNA array) **/
NeighboursN = InterpretArgs_ (NeighbourIA, SelectedResidue,
			      InFileName, OutFileName, MapFileName,
			      argc, argv);

/** Exit on failure: **/
if (NeighboursN < 0) exit (-1);

/** If there are no neighbour indices specified, use default: **/
if (NeighboursN == 0)
	{
	WarningMessage_ ("neigh", "main", "",
			 "No neighbours specified - default used",
			 " (a single neighbour at position -3).\n", "", "");
	NeighboursN = 1;
	NeighbourIA[0] = -3;
	}

/* Check the amino acid name: */
if (CheckAminoAcidName_ (namesSA, SelectedResidue) < 0) exit (-2);

/* Read map file and replace built-in numeric values, if map file specified: */
if (strlen (MapFileName) > 0)
	{
	n = ReadMapFile_ (valuesA, namesSA, MapFileName);
	if (n < 0) exit (-3);
	}

/* Do the real job: */
n = RawJob_ (NeighbourIA, NeighboursN, valuesA, namesSA,
	     SelectedResidue, InFileName, OutFileName);
if (n < 0) exit (-3);

/* The end: */
return 1;
}

/*======usage info:==========================================================*/

void PrintUsage_ (void)
{
fprintf (stdout,
   "\n==========================================================\n");
fprintf (stdout,
   "Find amino acid neighbour(s) at given relative position(s)\n");
fprintf (stdout,
   "==========================================================\n\n");
fprintf (stdout,
   "Usage:  neigh name infile outfile\n");
fprintf (stdout,
   "        neigh name [i1 i2 ... in] infile outfile [mapfile]\n");
fprintf (stdout,
   "        neigh -h                            ... print help\n\n");
fprintf (stdout,
   "name          ... amino acid name (three letters: ARG etc)\n");
fprintf (stdout,
   "i1            ... relative position of neighbour number 1.\n");
fprintf (stdout,
   "                  If not  specified, -3  used as  default.\n");
fprintf (stdout,
   "i2            ... relative position of neighbour number 2.\n");
fprintf (stdout,
   "                  Not required; no default.\n");
fprintf (stdout,
   "in            ... neighbour number n (may be omitted too).\n");
fprintf (stdout,
   "infile        ... input  file  name  (must  be specified).\n");
fprintf (stdout,
   "outfile       ... output file  name  (must  be specified).\n");
fprintf (stdout,
   "mapfile       ... file  with numeric values associated  to\n");
fprintf (stdout,
   "                  amino acids.  If not  specified, default\n");
fprintf (stdout,
   "                  scale  (Kyte-Doolittle hydrophob.) used.\n");
}

/*======detailed information:================================================*/

void PrintHelp_ (void)
{
fprintf (stdout,
"=========================================================================\n");
fprintf (stdout,
"    Read STRIDE output, look for each  occurence of selected  amino  acid\n");
fprintf (stdout,
"and identify  neighbours at specified  relative  positions.  Normally, at\n");
fprintf (stdout,
"least one relative position  should be specified.  If missing, -3 will be\n");
fprintf (stdout,
"used as  default and only  one neighbour  will be  listed in  the output.\n"); 
fprintf (stdout,
"    This example  explains  what  are \"neighbours at  specified  relative\n");
fprintf (stdout,
"positions\" (one letter code instead of  STRIDE data used for simplicity):\n");
fprintf (stdout,
"Neighbours of  THR  at positions  -3, 4, 3 and 4 in the sequence fragment\n");
fprintf (stdout,
"LLEGGFWQAITVCALGAF are: Q at -3, W at -4, A at 3 and L at 4.\n");
fprintf (stdout,
"    When the amino acid  at specified  position is  recognized, a numeric\n");
fprintf (stdout,
"value is assigned  to it. Numeric values are read  from map file.  If map\n");
fprintf (stdout,
"file is not specified, default scale  is used (Kyte-Doolitle hydrophob.).\n");
fprintf (stdout,
"    If n neighbours were requested, the output format will be:\n");
fprintf (stdout,
"(1) Numeric values  assigned to  neighbours (n numbers, floating  point).\n");
fprintf (stdout,
"(2) Amino acid name and relative position, for each neighbour.\n");
fprintf (stdout,
"(3) Position of  specified amino acid  with respect to the  nearest helix\n");
fprintf (stdout,
"    start. Positive if after h. start, neg. otherwise; +-99  if too  far.\n");
fprintf (stdout,
"(4) The position with respect to the nearest  helix end  (+ means after).\n");
fprintf (stdout,
"(5) Serial number, chain ID and second. struct. for specified amino acid.\n");
fprintf (stdout,
"-------------------------------------------------------------------------\n");
fprintf (stdout,
"Type       neigh       without  any  arguments,  to  obtain  usage  info.\n");
fprintf (stdout,
"=========================================================================\n");
}

/*======print error message:=================================================*/

/* Note: there is no  newline  in the last  four fprintf! */
/* You are expected to supply them through the arguments! */

void ErrorMessage_ (char *module, char *function, char *filename,
		    char *s1, char *s2, char *s3, char *s4)
{
fprintf (stderr, "\nERROR>");
fprintf (stderr, " module: %s, function: %s", module, function);
if (strlen (filename) != 0) fprintf (stderr, ", file: %s", filename);
fprintf (stderr, "\n");
if (strlen (s1) != 0) fprintf (stderr, "%s", s1);
if (strlen (s2) != 0) fprintf (stderr, "%s", s2);
if (strlen (s3) != 0) fprintf (stderr, "%s", s3);
if (strlen (s4) != 0) fprintf (stderr, "%s", s4);
}

/*======print warning message:===============================================*/

/* Note: there is no  newline  in the last  four fprintf! */
/* You are expected to supply them through the arguments! */

void WarningMessage_ (char *module, char *function, char *filename,
		      char *s1, char *s2, char *s3, char *s4)
{
fprintf (stdout, "WARNING>");
fprintf (stdout, " module: %s, function: %s", module, function);
if (strlen (filename) != 0) fprintf (stdout, ", file: %s", filename);
fprintf (stdout, "\n");
if (strlen (s1) != 0) fprintf (stdout, "%s", s1);
if (strlen (s2) != 0) fprintf (stdout, "%s", s2);
if (strlen (s3) != 0) fprintf (stdout, "%s", s3);
if (strlen (s4) != 0) fprintf (stdout, "%s", s4);
}

/*======separate command line arguments:=====================================*/

int InterpretArgs_ (int *IndexP, char *SelectedResidue,
		    char *InFileName, char *OutFileName, char *MapFileName,
		    int argc, char *argv[])
{
int		NumArgsN = 0, OtherArgsN = 0;
int		i, n;

/* Initialize strings: */
*SelectedResidue = '\0';
*InFileName = '\0';
*OutFileName = '\0';
*MapFileName = '\0';

/* Check all command line arguments except the program name: */
for (i = 1; i < argc; i++)
	{
	/** Argument is integer? **/
	if (IsInteger_ (argv[i]))
		{
		if (sscanf (argv[i], "%d", &n) != 1)
			{
			ErrorMessage_ ("neigh", "SeparateArguments", "",
				       "Failed to extract relative",
				       " position;",
				       " check arguments!\n", "");
			return -1;
			}
		*(IndexP + NumArgsN) = n;
		NumArgsN++;
		}
	/** If argument is not integer, extract textual information: **/
	else
		{
		if (OtherArgsN == 0)
			{
			strncpy (SelectedResidue, argv[i], STRINGSIZE - 1);
			*(SelectedResidue + STRINGSIZE - 1) = '\0';
			}
		else if (OtherArgsN == 1)
			{
			strncpy (InFileName, argv[i], STRINGSIZE - 1);
			*(InFileName + STRINGSIZE - 1) = '\0';
			}
		else if (OtherArgsN == 2)
			{
			strncpy (OutFileName, argv[i], STRINGSIZE - 1);
			*(OutFileName + STRINGSIZE - 1) = '\0';
			}
		else if (OtherArgsN == 3)
			{
			strncpy (MapFileName, argv[i], STRINGSIZE - 1);
			*(MapFileName + STRINGSIZE - 1) = '\0';
			}
		OtherArgsN++;
		}
	}

/* Check the number of non-numeric arguments (at least 3 should be present): */
if (OtherArgsN < 3)
	{
	ErrorMessage_ ("neigh", "InterpretArgs", "",
		"One amino acid name and at least two file names",
		" should be specified!\n",
		"Check the command line arguments!\n", "");
	return -1;
	}

/* Are there too many non-numeric arguments? */
if (OtherArgsN > 4)
	{
	WarningMessage_ ("neigh", "InterpretArgs", "",
		"Too many arguments found",
		" (not counting numeric arguments)!\n",
		"Extra arguments ignored ...\n\n", "");
	}

/* Compare file names (all should be different): */
if (!strcmp (InFileName, OutFileName))
	{
	ErrorMessage_ ("neigh", "InterpretArgs", InFileName,
		"Input file and output file should",
		" have different names!\n", "", "");
	return -2;
	}
else if (!strcmp (InFileName, MapFileName))
	{
	ErrorMessage_ ("neigh", "InterpretArgs", InFileName,
		"Input file and map file should",
		" have different names!\n", "", "");
	return -3;
	}
else if (!strcmp (OutFileName, MapFileName))
	{
	ErrorMessage_ ("neigh", "InterpretArgs", OutFileName,
		"Output file and map file should",
		" have different names!\n", "", "");
	return -4;
	}

/* Print all arguments: */

/** Numeric arguments: **/
fprintf (stdout, "Numeric arguments (relative positions):");
for (i = 0; i < NumArgsN; i++) fprintf (stdout, " %d", *(IndexP + i));
fprintf (stdout, "\n");

/** Other arguments (amino acid name, file names): **/
fprintf (stdout, "Amino acid name: %s\n", SelectedResidue);
fprintf (stdout, "Input file: %s\n", InFileName);
fprintf (stdout, "Output file: %s\n", OutFileName);
fprintf (stdout, "Map file: %s\n", MapFileName);

return NumArgsN;
}

/*======string contains digits only?=========================================*/

int IsInteger_ (char *s)
{
int		n;

while ((n = *s++) != '\0') if (!isdigit (n) && (n != '-')) return 0;

/* If this point is reached, all characters are digits: */
return 1;
}

/*======check the specified amino acid name:=================================*/

int CheckAminoAcidName_ (AminoAcidNameS *namesSP, char *s)
{
int		i;

if (strlen (s) != 3)
	{
	ErrorMessage_ ("neigh", "CheckAminoAcidName", "",
		"Use three letters code to specify the amino acid name!\n",
		"", "", "");
	return -1;
	}

for (i = 0; i < 23; i++) if (!strcmp ((namesSP + i)->LongName, s)) return i;

/* If this points is reached, amino acid is not recognized: */
ErrorMessage_ ("neigh", "CheckAminoAcidName", "",
	"Amino acid name (", s, ") not recognized!\n", "");
return -1;
}

/*======read map file:=======================================================*/

int ReadMapFile_ (double *valuesP, AminoAcidNameS *namesSP, char *FileName)
{
FILE		*File;
char		Line[STRINGSIZE];
int		size;
int		n;
double		value;

if ((File = fopen (FileName, "r")) == NULL)
	{
	ErrorMessage_ ("seq2num", "ReadMapFile", FileName,
		"Failed to open map file!\n",
		"Check the file name and path.\n",
		"", "");
	return -1;
	}

/* Reset values: */
for (n = 0; n < 23; n++) *(valuesP + n) = 0.0;

/* Read line by line: */
size = sizeof (Line);
while (fgets (Line, size, File) != NULL)
	{
	/** Skip comments (line beginning with #): **/
	if (Line[0] == '#') continue;

	/** Remove newline, if present: **/
	n = strlen (Line);
	if (Line[n - 1] == '\n') Line[n - 1] = '\0';

	/** Skip empty lines: **/
	if (strlen (Line) == 0) continue;

	/** Scan all amino acid names: **/
	for (n = 0; n < 23; n++)
		{
		if (!strstr (Line, (namesSP + n)->LongName)) continue;
		ExtractDigits_ (Line);
		if (sscanf (Line, "%lf", &value) != 1) continue;
		*(valuesP + n) = value;
		break;
		}
	}

fclose (File);
return 1;
}

/*======extract digits, minus signs and decimal points:======================*/

int ExtractDigits_ (char *s)
{
int		count = 0;
char		*p;
int		n;

p = s;
while ((n = *p++) != '\0')
	{
	if (isdigit (n)) continue;
	if (n == '-') continue;
	if (n == '.') continue;
	*(p - 1) = ' ';
	count++;
	}

/* Return the number of characters replaced: */
return count;
}

/*======do the real job:=====================================================*/

int RawJob_ (int *NeighbourIP, int NeighboursN, double *valuesP,
	     AminoAcidNameS *namesSP, char *SelectedResidue,
	     char *InFileName, char *OutFileName)
{
int		StructuresN = 0;
FILE		*InFile, *OutFile;
int		DataLeftF = 1;
ResidueDataS	*RDSP;
long		ResiduesN = 0;
long		MaxResiduesN = 10000, MemSize;

/* Open files: */
if (!(InFile = fopen (InFileName, "r")))
	{
	ErrorMessage_ ("neigh", "RawJob", InFileName,
		"Failed to open input file - check the specified file!\n",
		"", "", "");
	return -1;
	}
if (!(OutFile = fopen (OutFileName, "w")))
	{
	ErrorMessage_ ("neigh", "RawJob", OutFileName,
		"Failed to open output file - check the specified file!\n",
		"", "", "");
	fclose (InFile);
	return -2;
	}

/* Allocate memory: */
MemSize = MaxResiduesN * sizeof (ResidueDataS);
if (!(RDSP = (ResidueDataS *) malloc (MemSize)))
	{
	ErrorMessage_ ("neigh", "RawJob", "",
		"Failed to allocate memory!\n", "", "", "");
	return -3;
	}

/* Read file, structure by structure: */
while (DataLeftF)
	{
	/* Write structure header to output file: */
	fprintf (OutFile, "#Structure number %d:\n", StructuresN + 1);

	/* Read one structure; reallocate memory, if chain is really big: */
	/* (That's why  ReadStructure_  gets pointer to pointer to data!) */
	ResiduesN = ReadStructure_ (&DataLeftF, &RDSP, &MaxResiduesN, InFile);
	if (ResiduesN == 0) continue;

	/* Assign numeric value each amino acid: */
	AssignValues_ (RDSP, ResiduesN, namesSP, valuesP);

	/* Scan structure, print data to output file: */
	ScanStructure_ (OutFile,
		SelectedResidue, RDSP, ResiduesN, NeighbourIP, NeighboursN);
	
	/* If this point is reached, one more structure is successfuly read: */
	StructuresN++;
	}

/* Print the number of structures to stdout: */
printf ("Total number of structures: %d\n", StructuresN);

/* Close files: */
fclose (InFile);
fclose (OutFile);

/* Free memory: */
if (RDSP) free (RDSP);

/* Return positive value on success: */
return 1;
}

/*======read one structure:==================================================*/

long ReadStructure_ (int *DataLeftF, ResidueDataS **RDSPP, long *MaxResiduesNP,
		     FILE *InFile)
{
long		ResiduesN = 0;
char		Line[STRINGSIZE];
int		StructureReachedF = 0;
ResidueDataS	*CurrP;
char		*p;
long		MaxN, MemSize;
char		AuxString[STRINGSIZE];
long		SerialNumber;

/* Copy pointer: */
CurrP = *RDSPP;

/* Set DataLeftF; if EOF is reached, this will be unchanged: */
*DataLeftF = 0;

/* Prepare the maximal number of residues; if */
/* that number is reached, reallocate memory: */
MaxN = *MaxResiduesNP - 10;

/* Read file line by line; important data  should be in one block, with no */
/* comments inserted in this block. StructureReachedF takes care about it. */
while (fgets (Line, STRINGSIZE, InFile))
	{
	/* What kind of line is read? */
	if (strstr (Line, "ASG") != Line)
		{
		if (!StructureReachedF) continue;
		else
			{
			*DataLeftF = 1;
			break;
			}
		}
	else (StructureReachedF = 1);

	/* The first token is "ASG" - skip it: */
	if (!(p = strtok (Line, " \t\n"))) continue;

	/* The second token is amino acid name - store it: */
	if (!(p = strtok (NULL, " \t\n"))) continue;
	strncpy (CurrP->Name, p, 3);
	CurrP->Name[3] = '\0';

	/* The third token is chain identifier: */
	if (!(p = strtok (NULL, " \t\n"))) continue;
	strncpy (CurrP->ChainID, p, 1);
	CurrP->ChainID[2] = '\0';

	/* The fourth token is serial number: */
	if (!(p = strtok (NULL, " \t\n"))) continue;
	strncpy (AuxString, p, STRINGSIZE - 1);
	AuxString[STRINGSIZE - 1] = '\0';
	ExtractDigits_ (AuxString);
	if (sscanf (AuxString, "%ld", &SerialNumber) != 1) continue;
	CurrP->Serial = SerialNumber;

	/* Skip the fifth and sixth token: */
	if (!(p = strtok (NULL, " \t\n"))) continue;
	if (!(p = strtok (NULL, " \t\n"))) continue;
	
	/* The seventh token contains secondary structure: */
	if (!(p = strtok (NULL, " \t\n"))) continue;
	strncpy (CurrP->Structure, p, 19);
	CurrP->Structure[19] = '\0';

	/* Update pointer and counter: */
	CurrP++;
	ResiduesN++;

	/* Check is it necessary to reallocate memory: */
	if (ResiduesN >= MaxN - 10)
		{
		MaxN += 10000;
		*MaxResiduesNP += 10000;
		MemSize = *MaxResiduesNP * sizeof (ResidueDataS);
		*RDSPP = (ResidueDataS *) realloc (*RDSPP, MemSize);
		}
	}

return ResiduesN;
}

/*======assign numeric values to amino acids:================================*/

long AssignValues_ (ResidueDataS *RDSP,
		    long ResiduesN, AminoAcidNameS *namesSP, double *valuesP)
{
long		ResiduesRecogN = 0;
long		l;
ResidueDataS	*CurrP;
int		i;

for (l = 0; l < ResiduesN; l++)
	{
	CurrP = RDSP + l;
	for (i = 0; i < 23; i++)
		{
		if (strcmp (CurrP->Name, (namesSP + i)->LongName) == 0) break;
		}
	if (i <= 23)
		{
		CurrP->Value = *(valuesP + i);
		ResiduesRecogN++;
		}
	else 
		{
		CurrP->Value = 0.0;
		printf ("Amino acid not recognized: %s\n", CurrP->Name);
		}
	}

return ResiduesRecogN;
}

/*======scan structure, write requested data to output file:=================*/

long ScanStructure_ (FILE *OutFile,
		     char *SelectedResidue, ResidueDataS *RDSP, long ResiduesN,
		     int *NeighbourIP, int NeighboursN)
{
long		OutputLinesN = 0;
long		l0, l1;
ResidueDataS	*CurrP;
int		i, j;
int		InsideF, Distance;


/* Scan the whole structure: */
for (l0 = 0; l0 < ResiduesN; l0++)
	{
	CurrP = RDSP + l0;
	/* Check residue name: */
	if (strcmp (CurrP->Name, SelectedResidue) != 0) continue;

	/* If this point is reached, write output: */

	/** First write numeric values (or 99.99 if out of range): **/
	for (i = 0; i < NeighboursN; i++)
		{
		j = *(NeighbourIP + i);
		l1 = l0 + j;
		CurrP = RDSP + l1;
		if ((l1 >= 0) && (l1 < ResiduesN))
			{
			if ((CurrP->Value < 100.0) && (CurrP->Value > -100))
				fprintf (OutFile, " %6.2f", CurrP->Value);
			else fprintf (OutFile, "%.2f", CurrP->Value);
			}
		else fprintf (OutFile, "  99.99");
		}

	/** Write names and relative positions of requested neigbours: **/
	for (i = 0; i < NeighboursN; i++)
		{
		j = *(NeighbourIP + i);
		l1 = l0 + j;
		CurrP = RDSP + l1;
		if ((l1 >= 0) && (l1 < ResiduesN))
			{
			fprintf (OutFile, "  %s %+d", CurrP->Name, j);
			}
		else fprintf (OutFile, "  XXX 99");
		}

	/** Is selected residue inside helix (alpha or 3-10)? **/
	if (strstr ((RDSP + l0)->Structure, "Helix")) InsideF = 1;
	else InsideF = 0;

	/** Write distance from the nearest helix start: **/
	if (InsideF) Distance = LookBackward_ (RDSP, ResiduesN, l0, InsideF);
	else Distance = LookForward_ (RDSP, ResiduesN, l0, InsideF);
	fprintf (OutFile, "  S %+3d", Distance);

	/** Write distance from the nearest helix end: **/
	if (InsideF) Distance = LookForward_ (RDSP, ResiduesN, l0, InsideF);
	else LookBackward_ (RDSP, ResiduesN, l0, InsideF);
	fprintf (OutFile, " E %+3d", Distance);

	/** Finally  write  serial number, chain **/
	/** ID, secondary structure and newline: **/
	CurrP = RDSP + l0;
	if (CurrP->Serial < 9999) fprintf (OutFile, " %4ld", CurrP->Serial);
	else fprintf (OutFile, " %ld", CurrP->Serial);
	fprintf (OutFile, " %s  %s\n", CurrP->ChainID, CurrP->Structure);
	}


return OutputLinesN;
}

/*======find helix start:====================================================*/
int LookForward_ (ResidueDataS *RDSP, long ResiduesN, long l, int InsideF)
{
int		Distance = 0;
int		OldInsideF;

OldInsideF = InsideF;
while (l++ < ResiduesN)
	{
	Distance++;
	if (strstr ((RDSP + l)->Structure, "Helix")) InsideF = 1;
	else (InsideF = 0);
	if (InsideF != OldInsideF) break;
	if (Distance > 7) return 99;
	}

return Distance;
}

/*======find helix end:======================================================*/
int LookBackward_ (ResidueDataS *RDSP, long ResiduesN, long l, int InsideF)
{
int		Distance = 0;
int		OldInsideF;

OldInsideF = InsideF;
while (l-- >= 0)
	{
	Distance--;
	if (strstr ((RDSP + l)->Structure, "Helix")) InsideF = 1;
	else (InsideF = 0);
	if (InsideF != OldInsideF) break;
	if (Distance < -7) return -99;
	}

return Distance;
}

/*===========================================================================*/


