/*=============================================================================

				   ran2num.c

	    Assign numeric values to specified amino acid ranges.

Purpose: Create the file which will contain amino acid serial numbers and
	 numeric values assigned  according to the  secondary  structure.
	 Secondary structure assignments  are read from input file, which
	 should contain  lines consisting of three  tokens:  range start,
	 range end and numeric value assigned to this range.

Revision history:
(1)		13.12.1998	created

========includes:============================================================*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>

/*======defines:=============================================================*/

#define STRINGSIZE 1000

/*======function prototypes:=================================================*/

void		PrintUsage_ (void);
void		PrintHelp_ (void);
void		ErrorMessage_ (char *, char *, char *,
			       char *, char *, char *, char *);
int		InterpretArgs_ (double *, int *, int *, char *, char *,
				int, char **);
int		OverallRange_ (int *, int *, char *);
int		RawJob_ (char *, char *, double, int, int);
int		ParseLine_ (int *, int *, double *, char *);
int		IsPositive_ (char *s);

/*======main:================================================================*/

int main (int argc, char *argv[])
{
int		n;
int		FirstResidue = 0, LastResidue = 0;
double		DefaultValue = 0.0;
int		RangesN;
char		InFileName[STRINGSIZE], OutFileName[STRINGSIZE];

/* Check is help requested: */
for (n = 1; n < argc; n++)
	{
	if (strstr (argv[n], "-h") == argv[n])
		{
		PrintHelp_ ();
		exit (0);
		}
	}

/* At least three command line arguments expected: */
if (argc < 3)
	{
	fprintf (stderr,
		"\nAt least three command line arguments expected!\n");
	PrintUsage_ ();
	exit (-1);
	}

/* Warn user if too many arguments specified: */
if (argc > 6)
	{
	printf ("\nNo more than 6 command line arguments expected,");
	printf (" but %d found.\n", argc);
	n = argc - 6;
	argc = 6;
	if (n == 1) printf ("The last argument ignored ...\n");
	else printf ("The last %d arguments ignored ...\n", n);
	}

/* Interpret command line arguments: */
n = InterpretArgs_ (&DefaultValue, &FirstResidue, &LastResidue,
		    InFileName, OutFileName, argc, argv);
if (n < 0) exit (-2);

/* Read input file, write the output file: */
RangesN = RawJob_ (OutFileName,
		   InFileName, DefaultValue, FirstResidue, LastResidue);
if (RangesN < 0) exit (-3);
printf ("%d ranges read from input file.\n", RangesN);

return 1;
}

/*======usage info:==========================================================*/

void PrintUsage_ (void)
{
fprintf (stdout, "\n=====================================================\n");
fprintf (stdout, "Assign numeric values to specified amino acid ranges.\n");
fprintf (stdout, "=====================================================\n\n");
fprintf (stdout, "Usage:        ran2num n1:n2 [-d value] infile outfile\n");
fprintf (stdout, "              ran2num -h               ... print help\n\n");
fprintf (stdout, "n1:n2    ... the overall range. It corresponds to the\n");
fprintf (stdout, "             entire sequence; thus, typically n1 = 1,\n");
fprintf (stdout, "             n2 = number of amino acids in  sequence.\n");
fprintf (stdout, "value    ... default numeric value. It is assigned to\n");
fprintf (stdout, "             amino acid ranges not specified in input\n");
fprintf (stdout, "             file. If not specified, zero is assumed.\n");
fprintf (stdout, "infile   ... input file name. Valid lines contain the\n");
fprintf (stdout, "             first and last amino acid  serial number\n");
fprintf (stdout, "             for a  given  range,  and numeric  value\n");
fprintf (stdout, "             which should be assigned  to this range.\n");
fprintf (stdout, "             Example:    102 125 1.0\n");
fprintf (stdout, "outfile  ... output file name.\n");
}

/*======detailed information:================================================*/

void PrintHelp_ (void)
{
fprintf (stdout,
   "==================================================================\n");
fprintf (stdout,
   "The purpose of this program is to prepare the file with amino acid\n");
fprintf (stdout,
   "serial numbers and numeric values  associated to amino acids. Such\n");
fprintf (stdout,
   "file may be used to present the secondary structure in a graphical\n");
fprintf (stdout,
   "form, using gnuplot or some other  plotting  program. For example,\n");
fprintf (stdout,
   "if input file contains the line:\n");
fprintf (stdout, "    3    6   1.0\n");
fprintf (stdout, " ... the first eight lines in the output file will be:\n");
fprintf (stdout, "    1   0.00\n    2   0.00\n    3   1.00\n    4   1.00\n");  
fprintf (stdout, "    5   1.00\n    6   1.00\n    7   0.00\n    8   0.00\n");
fprintf (stdout,
   "All amino acids not covered by  specified  ranges will be assigned\n");
fprintf (stdout,
   "default numeric value. If flag -d is not used, default value  will\n");
fprintf (stdout,
   "be equal to zero.\n");
fprintf (stdout,
   "------------------------------------------------------------------\n");
fprintf (stdout,
   "Type      ran2num     without any arguments, to obtain usage info.\n");
fprintf (stdout,
   "==================================================================\n");
}

/*======print error message:=================================================*/

/* Note: there is no  newline  in the last  four fprintf! */
/* You are expected to supply them through the arguments! */

void ErrorMessage_ (char *module, char *function, char *filename,
		    char *s1, char *s2, char *s3, char *s4)
{
fprintf (stderr, "\nERROR>");
fprintf (stderr, " module: %s, function: %s", module, function);
if (strlen (filename) != 0) fprintf (stderr, ", file: %s", filename);
fprintf (stderr, "\n");
if (strlen (s1) != 0) fprintf (stderr, "%s", s1);
if (strlen (s2) != 0) fprintf (stderr, "%s", s2);
if (strlen (s3) != 0) fprintf (stderr, "%s", s3);
if (strlen (s4) != 0) fprintf (stderr, "%s", s4);
}

/*======interpret command line arguments:====================================*/

int InterpretArgs_ (double *DefaultValue,
		    int *FirstResidueP, int *LastResidueP,
		    char *InFileName, char *OutFileName,
		    int argc, char **argv)
{
int		i;
double		v;

/* Initialize strings: */
InFileName[0] = '\0';
OutFileName[0] = '\0';

/* Scan all arguments: */
for (i = 1; i < argc; i++)
	{
	/* Overall range specification? */
	if (OverallRange_ (FirstResidueP, LastResidueP, argv[i]) > 0)
		{
		printf ("Overall range: from %d to %d\n",
			*FirstResidueP, *LastResidueP);
		continue;
		}

	/* Default value redefinition? */
	if (strcmp (argv[i], "-d") == 0)
		{
		i++;
		if (sscanf (argv[i], "%lf", &v) != 1)
			{
			ErrorMessage_ ("ran2num", "InterpretArgs", "",
				"Failed to read default numeric value!",
				" Flag -d used, but value not specified.\n",
				"", "");
			return -1;
			}
		*DefaultValue = v;
		printf ("Default numeric value: %.2f\n", *DefaultValue);
		continue;
		}

	/* Input file name: */
	if (strlen (InFileName) == 0)
		{
		strncpy (InFileName, argv[i], STRINGSIZE - 1);
		InFileName[STRINGSIZE - 1] = '\0';
		printf ("Input file: %s\n", InFileName);
		continue;
		}

	/* Output file name: */
	if (strlen (OutFileName) == 0)
		{
		strncpy (OutFileName, argv[i], STRINGSIZE - 1);
		OutFileName[STRINGSIZE - 1] = '\0';
		printf ("Output file: %s\n", OutFileName);
		continue;
		}

	i++;
	}

/* Return positive on success: */
return 1;
}

/*======try to interpret argument as overall range specification:============*/

int OverallRange_ (int *FirstResidueP, int *LastResidueP, char *s)
{
char		*p;
int		n1, n2;

/* Scan string: */
p = s;
while ((n1 = *p++) != '\0') if (!isdigit (n1) && (n1 != ':')) return -1;

/* If this point is reached, string contains overall range specification: */
/** Read the first token: **/
if ((p = strtok (s, ":")) == NULL) return -2;
if (sscanf (p, "%d", &n1) != 1) return -3;
if (n1 < 0) return -4;

/** Read the second token: **/
if ((p = strtok (NULL, " :\t\n")) == NULL) return -5;
if (sscanf (p, "%d", &n2) != 1) return -6;
if (n2 < 0) return -7;
if (n2 < n1) return -8;

/* If this point is reached, range specification is successfuly read: */
*FirstResidueP = n1;
*LastResidueP = n2;

return 1;
}

/*======do the real job:=====================================================*/

int RawJob_ (char *OutFileName, char *InFileName, double DefaultValue,
	     int FirstResidue, int LastResidue)
{
int		RangesN = 0;
FILE		*InFile, *OutFile;
char		Line[STRINGSIZE];
int		LineSize;
int		RangeStart, OldRangeStart, RangeEnd, GapStart, GapEnd;
double		Value;
int		i;

LineSize = sizeof (Line);

/* Open input file (read only): */
if ((InFile = fopen (InFileName, "r")) == NULL)
	{
	ErrorMessage_ ("ran2num", "RawJob", "InFileName",
		"Failed to open input file!\n", "", "", "");
	return -1;
	}

/* Open output file for write: */
if ((OutFile = fopen (OutFileName, "w")) == NULL)
	{
	ErrorMessage_ ("ran2num", "RawJob", "OutFileName",
		"Failed to open output file!\n", "", "", "");
	fclose (InFile);
	return -2;
	}

GapStart = FirstResidue;
OldRangeStart = 0;

/* Read input file, line by line: */
while (fgets (Line, LineSize, InFile))
	{
	/* Skip comments and empty lines: */
	if (strstr (Line, "#") == Line) continue;
	if (strlen (Line) <= 1) continue;

	/* Parse line: */
	if (ParseLine_ (&RangeStart, &RangeEnd, &Value, Line) < 0) continue;

	/* Check new indices: */
	if (RangeStart < OldRangeStart) continue;
	if (RangeEnd > LastResidue) RangeEnd = LastResidue;

	/* Increase count; */
	RangesN++;

	/* Write data to output file: */
	for (i = GapStart; i < RangeStart; i++)
		fprintf (OutFile, "%5d %8.2f\n", i, DefaultValue);
	for (i = RangeStart; i <= RangeEnd; i++)
		fprintf (OutFile, "%5d %8.2f\n", i, Value);

	/* Set new gap data: */
	GapStart = RangeEnd + 1;

	/* Store the range start: */
	OldRangeStart = RangeStart;
	}

/* Add the last gap: */
GapEnd = LastResidue;
for (i = GapStart; i <= GapEnd; i++)
	fprintf (OutFile, "%5d %8.2f\n", i, DefaultValue);

fclose (InFile);
fclose (OutFile);
return RangesN;
}

/*======parse line, extract three tokens:====================================*/

int ParseLine_ (int *StartP, int *EndP, double *ValueP, char *s)
{
char		*p;

/* Take the first token; check is it positive integer, and extract value: */
if ((p = strtok (s, " \t,;")) == NULL) return -1;
if (!IsPositive_ (p)) return -2;
if (sscanf (p, "%d", StartP) != 1) return -3;
if (*StartP < 0) return -4;

/* The second token: */
if ((p = strtok (NULL, " \t,;")) == NULL) return -5;
if (!IsPositive_ (p)) return -6;
if (sscanf (p, "%d", EndP) != 1) return -7;
if (*EndP < 0) return -8;
if (*EndP < *StartP) return -9;

/* The third token: */
if ((p = strtok (NULL, " \t,;")) == NULL) return -10;
if (sscanf (p, "%lf", ValueP) != 1) return -11;

/* If this point is reached, everything was read: */
return 1;
}

/*======check is string made of digits:======================================*/

int IsPositive_ (char *s)
{
char		*p;
int		n;

p = s;
while ((n = *p++) == '\0') if (!isdigit (n)) return 0;

return 1;
}

/*===========================================================================*/


