/*=============================================================================

				   relab.c
		Remove chain identifiers and reindex PDB file

==========includes:==========================================================*/

#include <stdio.h>
#include <string.h>

/*======defines:=============================================================*/

#define STRINGSIZE 1000

/*======function prototypes:=================================================*/

void		PrintUsage_ (void);
void		PrintHelp_ (void);
int		ExtractFileNames_ (char *, int, char *, int, int, char **);
int		RawJob_ (char *, char *);
char		*AminoAcidLabel_ (char *, char *);
int		WriteModifiedLine_ (FILE *, char *, int);

/*======main:================================================================*/

int main (int argc, char *argv[])
{
int		i;
char		InFileName[1000], OutFileName[1000];

/* Check is help requested: */
for (i = 1; i < argc; i++)
	{
	if (strstr (argv[i], "-h") == argv[i])
		{
		PrintHelp_ ();
		PrintUsage_ ();
		exit (0);
		}
	}

/* At least three command line arguments expected: */
if (argc < 3)
	{
	fprintf (stderr, "\nThree command line arguments expected!\n");
	PrintUsage_ ();
	exit (-1);
	}

/* Extract file names: */
if (ExtractFileNames_ (OutFileName, 999, InFileName, 999, argc, argv) < 2)
	{
	fprintf (stderr, "\nFailed to extract file names;");
	fprintf (stderr, " check command line arguments!\n");
	PrintUsage_ ();
	exit (-2);
	}

/* Do something: */
if (RawJob_ (OutFileName, InFileName) < 0) exit (-3);

/* Return positive number on success: */
return 1;
}

/*======usage info:==========================================================*/

void PrintUsage_ (void)
{
fprintf (stdout, "\n=============================================\n");
fprintf (stdout, "Remove chain identifiers and reindex PDB file\n");
fprintf (stdout, "=============================================\n\n");
fprintf (stdout, "Usage:         relab infile outfile\n");
fprintf (stdout, "               relab -h        ... print help\n\n");
}

/*======detailed information:================================================*/

void PrintHelp_ (void)
{
fprintf (stdout,
   "\n================================================================\n");
fprintf (stdout,
   "The purpose of this program is to replace  chain identifiers and\n");
fprintf (stdout,
   "chain amino acid serial numbers with unique serial numbers. This\n");
fprintf (stdout,
   "should help with rasmol version 2.5.1, to select unique range of\n");
fprintf (stdout,
   "amino acids. The program recognizes PDB version 2.1 file format.\n");
fprintf (stdout,
   "================================================================\n");
}

/*======extract file names:==================================================*/

int ExtractFileNames_ (char *OutFileName, int OutFileNameSize,
		       char *InFileName, int InFileNameSize,
		       int argc, char *argv[])
{
int		ReturnValue = 0;
int		i;

/* Scan command line arguments backwards: */
for (i = argc - 1; i > 0; i--)                /* Do not check program name! */
	{
	if (strstr (argv[i], "-") == argv[i]) continue;
	if (ReturnValue == 0)
		{
		strncpy (OutFileName, argv[i], OutFileNameSize - 1);
		OutFileName[OutFileNameSize - 1] = '\0';
		}
	else if (ReturnValue == 1)
		{
		strncpy (InFileName, argv[i], InFileNameSize - 1);
		InFileName[InFileNameSize - 1] = '\0';
		}
	else break;
	ReturnValue++;
	}

/* Compare file names (should be different): */
if (strcmp (OutFileName, InFileName) == 0)
	{
	fprintf (stderr, "\nInput file name and output file name should be");
	fprintf (stderr, " different!\n");
	return -1;
	}

/* Return the number of file names extracted (two if successful): */
return ReturnValue;
}

/*======do the raw job:======================================================*/

int RawJob_ (char *OutFileName, char *InFileName)
{
int		AminoAcidsN = 0;
FILE		*InFile, *OutFile;
char		Line[STRINGSIZE];
int		LineLength;
char		AtomID[5] = "ATOM";
char		NewLabel[20], OldLabel[20] = "XXXXXX";

/* Open input file: */
if ((InFile = fopen (InFileName, "r")) == NULL)
	{
	fprintf (stderr, "\nFailed to open input file!\n");
	return -1;
	}

/* Open output file: */
if ((OutFile = fopen (OutFileName, "w")) == NULL)
	{
	fprintf (stderr, "\nFailed to open output file!\n");
	fclose (InFile);
	return -1;
	}

/* Read line by line: */
LineLength = sizeof (Line);
while (fgets (Line, LineLength, InFile) != NULL)
        {
        /* Check the line (only ATOM records should be modified): */
        if (strstr (Line, AtomID) != Line)
		{
		fputs (Line, OutFile);
		continue;
		}

	/* Skip short lines: */
	if (strlen (Line) < 30) continue;		 /* Why 30? Why not! */

	/* Extract the amino acid label: */
	if (!AminoAcidLabel_ (NewLabel, Line)) continue;

	/* Check is it necessary to increase the amino acid serial number: */
	if (strcmp (OldLabel, NewLabel) != 0) AminoAcidsN++;

	/* Write output data to file: */
	WriteModifiedLine_ (OutFile, Line, AminoAcidsN);

	/* Copy label: */
	strncpy (OldLabel, NewLabel, 9);
	}

/* Total number of amino acids: */
printf ("\n%d amino acid(s) found.\n", AminoAcidsN);

/* Close files: */
fclose (InFile);
fclose (OutFile);

/* Return the number of amino acids: */
return AminoAcidsN;

}

/*======extract amino acid label (chain ID and number):======================*/

char *AminoAcidLabel_ (char *dest, char *src)
{
int		StartI = 21, Length = 5;
char		*p;

/* Copy substring: */
p = src + StartI;
strncpy (dest, p, Length);
dest[Length] = '\0';

/* Return value on success: */
return dest;
}

/*======write modified input line:===========================================*/

int WriteModifiedLine_ (FILE *OutFile, char *Line, int AminoAcidsN)
{
int		ChainIDPosition = 21, LastPartStart = 26;

/* Replace chain ID with zero: */
*(Line + ChainIDPosition) = '\0';

/* Write the first part (21 character): */
fprintf (OutFile, "%s", Line);

/* Write the new amino acid serial number preceded by space: */
/* (space replaces chain identifier) */
fprintf (OutFile, " %4d", AminoAcidsN);

/* Write the remainder of the line: */
fprintf (OutFile, "%s", Line + LastPartStart);

/* Return value has no meaningo at present: */
return 1;
}

/*===========================================================================*/


