/*=============================================================================

				rsc.c

			   remove side chain	

Remove side chains for all amino acids in pdb file.

Revision history:
(1)		created 11.3.1998

========includes:============================================================*/

#include <stdio.h>
#include <string.h>

/*======defines:=============================================================*/

#define MAXRANGES 20
#define MINLINELENGTH 2

/*======function prototypes:=================================================*/

void		PrintUsage (void);
int		CopyFileNames (char *, int, char *, int, int, char **);
int		CopyLines (char *, char *);
int		RenameFile (char *, char *, int);
int		RemoveNewline (char *);
int		CheckLine (char *, char *);
int		ExtractAtomName (char *, char *, char *);
void		ErrorMessage (char *, char *, char *,
			      char *, char *, char *, char *);
void		WarningMessage (char *, char *, char *,
				char *, char *, char *, char *);

/*===========================================================================*/

int main (int argc, char *argv[])
{
char		InFileName[1000], OutFileName[1000];
int		n1, n2;


/* Three command line arguments are expected: */
if (argc != 3)
	{
	PrintUsage ();
	exit (0);
	}

/* Prepare file names: */
n1 = sizeof (InFileName) - 1;
n2 = sizeof (OutFileName) - 1;
if (CopyFileNames (InFileName, n1, OutFileName, n2, argc, argv) < 0) exit (0);
if (CopyLines (InFileName, OutFileName) < 0) exit (0);

/* The end: */
return 1;
}

/*======print usage info:====================================================*/

void PrintUsage (void)
{
fprintf (stdout, "\n=============================\n");
fprintf (stdout, "Remove amino acid side chains\n");
fprintf (stdout, "=============================\n\n");
fprintf (stderr, "Usage:     rsc infile outfile\n\n");
}

/*======copy file names (the last two command line arguments):===============*/

int CopyFileNames (char *InFileName,  int IFNMaxChars,
		   char *OutFileName, int OFNMaxChars,
		   int argc, char **argv)
{
/* Copy the input and output file name: */
strncpy (InFileName,  argv[argc - 2], IFNMaxChars);
InFileName[IFNMaxChars] = '\0';
strncpy (OutFileName, argv[argc - 1], OFNMaxChars);
OutFileName[IFNMaxChars] = '\0';

/* Compare file names (should be different): */
if (strcmp (InFileName, OutFileName) == 0)
	{
	ErrorMessage ("rsc", "CopyFileNames", InFileName,
		      "The same file name is specified",
		      " for both input and output file.\n", "", "");
	return -1;
	}

/* Print file names: */
fprintf (stdout, "Input file: %s, output file: %s\n", InFileName, OutFileName);

return 1;
}

/*======copy lines:==========================================================*/

int CopyLines (char *InFileName, char *OutFileName)
{
FILE		*InFile = NULL, *OutFile = NULL;
char		BackupFileName[1000];
int		MaxChars;
int		n;
char		Line[1000];

/* Check the input file: */
if ((InFile = fopen (InFileName, "r")) == NULL)
	{
	ErrorMessage ("rsc", "CopyLines", InFileName,
			"Failed to open the input file.",
			" Check the specified pdb file!\n", "", "");
	return -1;
	}

/* Check the output file (exists allready?): */
if ((OutFile = fopen (OutFileName, "r")) != NULL)
	{
	fclose (OutFile);
	OutFile = NULL;
	MaxChars = (int) sizeof (BackupFileName);
	n = RenameFile (BackupFileName, OutFileName, MaxChars);
	if (n > 0) WarningMessage ("rsc", "CopyLines", OutFileName,
			"The specified output file exists. It will be",
			" renamed to ", BackupFileName, "\n");
	}

if ((OutFile = fopen (OutFileName, "w")) == NULL)
	{
	ErrorMessage ("rsc", "CopyLines", OutFileName,
		"Failed to open the output file.",
		" Check who is the owner of this file;\n",
		"check directory, protection",
		" and available disk space.\n");
	if (InFile) fclose (InFile);
	return -1;
	}

/* Read the input file, line by line: */
while (fgets (Line, MaxChars, InFile) != NULL)
	{
	/* Remove newline, if present: */
	RemoveNewline (Line);
	/* Skip line, if too short: */
	if (strlen (Line) <= 1) continue;
	/* Process this line: */
	if (CheckLine (Line, OutFileName) < 0) continue;
	fprintf (OutFile, "%s\n", Line);
	}

/* Close files: */
fclose (InFile);
fclose (OutFile);

return 1;
}

/*======rename file (put # before the old file name):========================*/

int RenameFile (char *BackupName, char *OldName, int FileNameLen)
{
int		SpaceLeft;

strcpy (BackupName, "#");
SpaceLeft = FileNameLen - strlen (BackupName) - 1;
if (SpaceLeft <= 0) return -1;
strncat (BackupName, OldName, SpaceLeft);
if (rename (OldName, BackupName) != 0) return -1;

return 1;
}

/*======remove newline, if present:==========================================*/

int RemoveNewline (char *Line)
{
int		LastCharPos;

LastCharPos = strlen (Line) - 1;
if (Line[LastCharPos] == '\n')
	{
	Line[LastCharPos] = '\0'; 
	return 1;
	}

return 0;
}

/*======check line:==========================================================*/

int CheckLine (char *Line, char *FileName)
{
char		Atom[5] = "ATOM";
int		i;
char		MainChainAtom[5][3] = {"N ", "CA", "C ", "O ", "CB"};

/* Return 1 if line does not contain atomic coordinates: */
if (strstr (Line, Atom) != Line) return 1;

/* Return  1 if line  does refer to atomic coordinates,  but atom */
/* is either N, CA, O or CB (i.e. does not belong to side chain): */
if (ExtractAtomName (Atom, Line, FileName) < 0) return -1;
for (i = 0; i < 5; i++) if (strcmp (Atom, MainChainAtom[i]) == 0) return 1;

return -1;
}

/*======extract atom ID (name):==============================================*/

int ExtractAtomName (char *Atom, char *Line, char *FileName)
{
int		AtomNameOffset = 13;

/* Check line length: */
if (strlen (Line) < 16)
	{
	WarningMessage ("rsc", "ExtractAtomName", FileName,
			"Line to  short.",
			" (without newline less than 15 characters):\n",
			Line, "\n");
	return -1;
	}

/* Copy two characters: */
Atom[0] = Line[AtomNameOffset];
Atom[1] = Line[AtomNameOffset + 1];
Atom[2] = '\0';

return 1;
}

/*======print error message:=================================================*/

/* Note: there is no  newline  in the last  four fprintf! */
/* You are expected to supply them through the arguments! */

void ErrorMessage (char *module, char *function, char *filename,
		   char *s1, char *s2, char *s3, char *s4)
{
fprintf (stderr, "\nERROR>");
fprintf (stderr, " module: %s, function: %s", module, function);
if (strlen (filename) != 0) fprintf (stderr, ", file: %s", filename);
fprintf (stderr, "\n");
if (strlen (s1) != 0) fprintf (stderr, "%s", s1);
if (strlen (s2) != 0) fprintf (stderr, "%s", s2);
if (strlen (s3) != 0) fprintf (stderr, "%s", s3);
if (strlen (s4) != 0) fprintf (stderr, "%s", s4);
}

/*======print warning message:===============================================*/

/* Note: there is no  newline  in the last  four fprintf! */
/* You are expected to supply them through the arguments! */

void WarningMessage (char *module, char *function, char *filename,
		     char *s1, char *s2, char *s3, char *s4)
{
fprintf (stdout, "WARNING>");
fprintf (stdout, " module: %s, function: %s", module, function);
if (strlen (filename) != 0) fprintf (stdout, ", file: %s", filename);
fprintf (stdout, "\n");
if (strlen (s1) != 0) fprintf (stdout, "%s", s1);
if (strlen (s2) != 0) fprintf (stdout, "%s", s2);
if (strlen (s3) != 0) fprintf (stdout, "%s", s3);
if (strlen (s4) != 0) fprintf (stdout, "%s", s4);
}

/*===========================================================================*/


