/*=============================================================================

				  rtf.c

			replace temperature factors

Purpose: replace temperature factors in PDB file, to assign different
colors on amino acids with different charges (in rasmol). 

Revision history:
(1)		created 11.2.1998

==========includes:==========================================================*/

#include <stdio.h>
#include <string.h>
#include "rtfdefs.h"

/*======defines:=============================================================*/

#define STRINGSIZE 1000

/*======function prototypes:=================================================*/

void		PrintUsage (void);
void		PrintHelp (void);
int		LoadTempFactors (TempFactorS *, int, char **);
int		ReadTempFactors (TempFactorS *, char *);
float		ReadTFValue (float *, char *, char *);
void		ErrorMessage (char *, char *, char *,
			      char *, char *, char *, char *);
void		WarningMessage (char *, char *, char *,
				char *, char *, char *, char *);
int		PrepareFileNames (char *, int, char *, int, int, char **);
int		ReplaceTempFactors (TempFactorS *, char *, char *);
int		RenameFile (char *, char *, int);
int		ExtractAminoAcidName (char *, char *, char *);
int		FindTempFactor (double *, char *, TempFactorS *, char *);
int		MergeTempFactor (char *, double, char *);

/*======main:================================================================*/

int main (int argc, char *argv[])
{
int		i, n1, n2;
TempFactorS	TempFactorSA[23] =
		{{"ALA", 1.0, 0}, {"CYS", 1.0, 0}, {"ASP", 1.0, 0},
		 {"GLU", 1.0, 0}, {"PHE", 1.0, 0}, {"GLY", 1.0, 0},
		 {"HIS", 1.0, 0}, {"ILE", 1.0, 0}, {"LYS", 1.0, 0},
		 {"LEU", 1.0, 0}, {"MET", 1.0, 0}, {"ASN", 1.0, 0},
		 {"PRO", 1.0, 0}, {"GLN", 1.0, 0}, {"ARG", 1.0, 0},
		 {"SER", 1.0, 0}, {"THR", 1.0, 0}, {"VAL", 1.0, 0},
		 {"TRP", 1.0, 0}, {"TYR", 1.0, 0}, {"ASX", 1.0, 0},
		 {"GLX", 1.0, 0}, {"UNK", 1.0, 0}};
char		InFileName[STRINGSIZE], OutFileName[STRINGSIZE];

/* Check if help requested: */
for (i = 1; i < argc; i++)
	{
	if (strstr (argv[i], "-h") == argv[i])
		{
		PrintHelp ();
		PrintUsage ();
		exit (0);
		}
	}

/* At least three arguments are needed: */
if (argc < 3)
	{
	PrintUsage ();
	exit (0);
	}

/* If option -f used, replace the temperature factors: */
if (LoadTempFactors (TempFactorSA, argc, argv) < 0) exit (0);
	
/* Prepare the file names (input pdb file, output pdb file): */
n1 = sizeof (InFileName) - 1;
n2 = sizeof (OutFileName) - 1;
if (PrepareFileNames (InFileName, n1, OutFileName, n2, argc, argv) < 0) 
	{
	exit (0);
	}

/* Read pdb file, replace all temperature factors, write to file: */
if (ReplaceTempFactors (TempFactorSA, InFileName, OutFileName) < 0) exit (0);

/* The end: */
return 1;

}

/*======usage info:==========================================================*/

void PrintUsage (void)
{
fprintf (stdout, "\n===========================================\n");
fprintf (stdout, "Replace the temperature factors in pdb file\n");
fprintf (stdout, "===========================================\n\n");
fprintf (stdout, "Usage:     rtf infile outfile [-f datafile]\n");
fprintf (stdout, "           rtf -h   ... print detailed help\n\n");
fprintf (stdout, "If option  -f  is used, datafile  should be\n");
fprintf (stdout, "supplied.  This file  should contain  amino\n");
fprintf (stdout, "acid names and temperature factors, written\n");
fprintf (stdout, "in separate lines;  in the  example  below,\n");
fprintf (stdout, "temp. factors for  Arg and Lys are defined:\n");
fprintf (stdout, "ARG  70.0\nLYS  60.0\n");
}

/*======detailed information:================================================*/

void PrintHelp (void)
{
fprintf (stdout,
   "======================================================================\n");
fprintf (stdout,
   "The purpose of this program is the replacement of  temperature factors\n");
fprintf (stdout,
   "in a pdb file.  The temperature factor for a given  amino acid will be\n");
fprintf (stdout,
   "set according to the amino acid charge or hydrophobicity. The modified\n");
fprintf (stdout,
   "pdb file  may be  used for  graphical presentation of charge or hydro-\n");
fprintf (stdout,
   "phobicity distribution in the protein with a known structure.\n"); 
fprintf (stdout,
   "======================================================================\n");
}

/*======check is temperature factor replacement requested:===================*/

int LoadTempFactors (TempFactorS *TFSP, int argc, char **argv)
{
int		i, FileNamePosition = -1, n;


/* Check the last argument for -f flag: */
if (strstr (argv[argc - 1], "-f") == argv[argc - 1])
	{
	ErrorMessage ("rtf", "LoadTempFactors", "",
		"File name missing!\n",
		"You tryed to use -f option but failed to\n",
		"specify the name of the temperature factor file.\n", "");
	return -1;
	}

/* Check all arguments (except the last one) for -f flag: */
for (i = 1; i < argc - 1; i++)
        {
        if (strstr (argv[i], "-f") == argv[i]) FileNamePosition = i + 1; 
	}

/* If flag -f not found, return 0 to caller: */
if (FileNamePosition == -1) return 0;

/* Flag -f was found, check the number of command line arguments: */
if (argc != 5)
	{
	ErrorMessage ("rtf", "LoadTempFactors", "",
		"If using -f option, you have to specify",
		" three files: input pdb file,\noutput pdb",
		" file and file with temperature factors",	
		" (datafile)\n");
	return -1;
	}

/* Read and replace temperature factors: */
n = ReadTempFactors (TFSP, argv[FileNamePosition]);
if (n == -1) return -1;
else if (n == 0)
	{
	WarningMessage ("rtf", "main", "",
			"Strange, the file with temperature factors",
			" specified, but no temperature\n",
			"factors changed. Check the file format",
			" and contents!\n");
        }
else fprintf (stdout, "%d temperature factor(s) replaced.\n", n);

return n;
}

/*======read temperature factors from the specified file:====================*/

int             ReadTempFactors (TempFactorS *TFSP, char *FileName)
{
FILE		*InFile;
char		Line[STRINGSIZE];
int		LineLength, i, n;
float		TFValue;
int		TempFactorsRead = 0;

/* Open file: */
if ((InFile = fopen (FileName, "r")) == NULL)
	{
	ErrorMessage ("rtf", "ReadTempFactors", FileName,
		"Failed to open file!\n",
		"You used -f option; be sure to specify the name\n",
		"of the file with temperature factors correctly.\n", "");
	return -1;
	}

/* Read line by line: */
LineLength = sizeof (Line);
while (fgets (Line, LineLength, InFile) != NULL)
	{
	/** Skip comments (line beginning with #): **/
	if (Line[0] == '#') continue;
	/** Scan all amino acid names: **/
	for (i = 0; i < 23; i++)
		{
		/*** Try to recognize amino acid name: ***/
		if (strstr (Line, TFSP[i].AminoAcidName) == NULL) continue;
		/*** Try to read temperature factor: ***/
		n = ReadTFValue (&TFValue, Line, TFSP[i].AminoAcidName);
		if (n == -1) continue;
		/*** Set t. f. value of everything OK: ***/
		TFSP[i].TempFactor = TFValue;
		/*** Changed for the first time? ***/
		if (TFSP[i].IsChanged == 0) TempFactorsRead++;
		/*** This one changed at least once: ***/
		TFSP[i].IsChanged = 1;
		}
	}

/* Close file: */
fclose (InFile);

/* Return the number of replaced temperature factors: */
return TempFactorsRead;
}

/*======read temperature factor value from character string (line):==========*/

float ReadTFValue (float *TFP, char *LineP, char *NameP)
{
char		*p;

/* Prepare pointer: */
p = strstr (LineP, NameP);
if (p == NULL) return -1;
else p += strlen (NameP);

/* Read temperature factor value: */
if (sscanf (p, "%f", TFP) == EOF) return -1;

return 1;
}

/*======print error message:=================================================*/

/* Note: there is no  newline  in the last  four fprintf! */
/* You are expected to supply them through the arguments! */

void ErrorMessage (char *module, char *function, char *filename,
		   char *s1, char *s2, char *s3, char *s4)
{
fprintf (stderr, "\nERROR>");
fprintf (stderr, " module: %s, function: %s", module, function);
if (strlen (filename) != 0) fprintf (stderr, ", file: %s", filename);
fprintf (stderr, "\n");
if (strlen (s1) != 0) fprintf (stderr, "%s", s1);
if (strlen (s2) != 0) fprintf (stderr, "%s", s2);
if (strlen (s3) != 0) fprintf (stderr, "%s", s3);
if (strlen (s4) != 0) fprintf (stderr, "%s", s4);
}

/*======print warning message:===============================================*/

/* Note: there is no  newline  in the last  four fprintf! */
/* You are expected to supply them through the arguments! */

void WarningMessage (char *module, char *function, char *filename,
		     char *s1, char *s2, char *s3, char *s4)
{
fprintf (stdout, "WARNING>");
fprintf (stdout, " module: %s, function: %s", module, function);
if (strlen (filename) != 0) fprintf (stdout, ", file: %s", filename);
fprintf (stdout, "\n");
if (strlen (s1) != 0) fprintf (stdout, "%s", s1);
if (strlen (s2) != 0) fprintf (stdout, "%s", s2);
if (strlen (s3) != 0) fprintf (stdout, "%s", s3);
if (strlen (s4) != 0) fprintf (stdout, "%s", s4);
}

/*======prepare file names (input and output pdb file):======================*/

int PrepareFileNames (char *InFileName,  int IFNMaxChars,
		      char *OutFileName, int OFNMaxChars,
		      int argc, char **argv)

{
int		i, j, MaxChars;
char 		FileName[2][STRINGSIZE];

/* Copy arguments; omit -f flag and the argument which follows that flag: */
j = 0;
i = 1;
MaxChars = sizeof (FileName[0]) - 1;
while (i < argc)
	{
	if (strstr (argv[i], "-f") == argv[i]) i += 2;
	else
		{
		strncpy (FileName[j], argv[i], MaxChars);
		FileName[j][MaxChars] = '\0';
		j++;
		if (j == 2) break;
		i++;
		}
	}

/* Check the number of copied arguments: */
if (j < 2)
	{
	ErrorMessage ("rtf", "PrepareFileNames", "",
		"Not enough command line arguments.\n",
		"If using -f flag, you have to specify three files.\n",
		"", "");
	return -1;
	}

/* Copy the input and output file name: */
strncpy (InFileName,  FileName[0], IFNMaxChars);
InFileName[IFNMaxChars] = '\0';
strncpy (OutFileName, FileName[1], OFNMaxChars);
OutFileName[OFNMaxChars] = '\0';

/* Compare file names (should be different): */
if (strcmp (InFileName, OutFileName) == 0)
	{
	ErrorMessage ("rtf", "PrepareFileNames", InFileName,
		"The same file name is specified",
		" for both input and output file.\n", "", "");
	return -1;
	}

/* Print file names: */
fprintf (stdout, "Input file: %s, output file: %s\n", InFileName, OutFileName);

return 1;
}

/*======Do the real job - replace t. f. in pdb file:=========================*/ 

int ReplaceTempFactors (TempFactorS *TFSP,
			char *InFileName, char *OutFileName)
{
FILE		*InFile = NULL, *OutFile = NULL;
char		BackupFileName[STRINGSIZE];
int		MaxChars;
int		n;
char		Line[STRINGSIZE];
char		AminoAcidName[30];
int		DataLineF, AminoAcidRecogF, TFMergedF;
double		TempFactor;
int		AllDoneF = 0, FailF = 0;

/* Check the input file: */
if ((InFile = fopen (InFileName, "r")) == NULL)
	{
	ErrorMessage ("rtf", "ReplaceTempFactors", InFileName,
		"Failed to open the input file.",
		" Check the specified pdb file!\n", "", "");
	return -1;
	}

/* Check the output file (exists allready?): */
if ((OutFile = fopen (OutFileName, "r")) != NULL)
	{
	fclose (OutFile);
	OutFile = NULL;
	MaxChars = (int) sizeof (BackupFileName);
	n = RenameFile (BackupFileName, OutFileName, MaxChars);
	if (n > 0) WarningMessage ("rtf", "ReplaceTempFactors", OutFileName,
			"The specified output file exists. It will be",
			" renamed to ", BackupFileName, "\n");
	}

/* Line length: */
MaxChars = sizeof (Line);

/* Formal, single pass do-while loop: */
do
	{
	/* Open the output file: */
	if ((OutFile = fopen (OutFileName, "w")) == NULL)
		{ 
		ErrorMessage ("rtf", "ReplaceTempFactors", OutFileName,
			"Failed to open the output file.",
			" Check who is the owner of this file;\n",
			"check directory, protection",
			" and available disk space.\n"); 
		break;
		}

	/* Read line by line: */
	while (fgets (Line, MaxChars, InFile) != NULL)
		{
		/* Reset flags: */
		DataLineF = 0;
		AminoAcidRecogF = 0;
		TFMergedF = 0;	

		/* Identify amino-acid: */
		DataLineF = ExtractAminoAcidName (AminoAcidName, Line,
						  InFileName);

		/* If line contains coordinates, find corresp. t. f.: */
		if (DataLineF > 0)
			{
			AminoAcidRecogF = FindTempFactor (&TempFactor,
					AminoAcidName, TFSP, InFileName);
			}

		/* Replace the temperature factor in given line: */
		if (AminoAcidRecogF > 0)
			{
			TFMergedF = MergeTempFactor (Line, TempFactor,
						     InFileName);
			}

		/* Write line: */
		if (fputs (Line, OutFile) == EOF)
			{
			FailF = 1;
			ErrorMessage ("rtf", "ReplaceTempFactors", OutFileName,
				      "Write failed; check the available",
				      " disk space!\n", "", "");
			break;
			}
		}
	if (FailF == 1) break;

	/* Set flag to notify that everything worked fine: */
	AllDoneF = 1;
	} while (0);

/* Close files: */
if (InFile) fclose (InFile);
if (OutFile) fclose (OutFile);

/* Return -1 on failure: */
if (AllDoneF == 0) return -1;

return 1;
}

/*======rename file (put # before the old file name):========================*/

int RenameFile (char *BackupName, char *OldName, int FileNameLen)
{
int		SpaceLeft;

strcpy (BackupName, "#");
SpaceLeft = FileNameLen - strlen (BackupName) - 1;
if (SpaceLeft <= 0) return -1;
strncat (BackupName, OldName, SpaceLeft);
if (rename (OldName, BackupName) != 0) return -1;

return 1;
}

/*======recognize amino acid (by name):======================================*/

int ExtractAminoAcidName (char *Name, char *Line, char *FileName)
{
char		AtomID[] = "ATOM";
int		AminoAcidOffset = 17;

/* Does this line refer to atomic coordinates? */
if (strstr (Line, AtomID) != Line) return 0;

/* Check the line length (must be 67 characters at least): */
if (strlen (Line) < 67)
	{
	WarningMessage ("rtf", "RecogAminoAcid", FileName,
			"Line too short",
			" (without newline less than 66 characters):\n",
			Line, "\n");
	return -1;
	}

/* Copy the amino acid name: */
if (strncpy (Name, Line + AminoAcidOffset, 3) == NULL) return -1;
Name[3] = '\0';

return 1;
}

/*======find the temperature factor for a given amino acid type:=============*/

int FindTempFactor (double *TempFactorP, char *Name,
		    TempFactorS *TFSP, char *FileName)
{
int		i, Flag = -1;

/* Scan all names, set temperature factor and flag if name recognized: */
for (i = 0; i < 23; i++)
	{
	if (strcmp (TFSP[i].AminoAcidName, Name) == 0)
		{
		*TempFactorP = TFSP[i].TempFactor;
		Flag = 1;
		break;
		}
	}

/* Name recognized? */
if (Flag < 0) WarningMessage ("rtf", "FindTempFactor", FileName,
			      "Amino acid not recognized: ", Name, "\n", "");

return Flag;
}

/*======merge temperature factor:============================================*/

int MergeTempFactor (char *Line, double TempFactor, char *FileName)
{
int		TempFactorStart = 60, TempFactorEnd = 65;
char 		InsertionString[100];
int		i, j;

if ((TempFactor > 999.99) || (TempFactor < -99.99))
	{
	sprintf (InsertionString, "%10.2f\n", TempFactor);
	ErrorMessage ("rtf", "MergeTempFactor", FileName,
			"Temperature factor out of range.",
			" The range is: -99.99 <= TempFactor <= 999.99\n",
			"Bad temperature factor value: ", InsertionString);
	return -1;
	}

/* Prepare the insertion string: */
sprintf (InsertionString, "%6.2f", TempFactor);

/* Check the line length: */
if (strlen (Line) < TempFactorEnd + 2)
	{
	ErrorMessage ("rtf", "MergeTempFactor", FileName,
			"Line too short (must be at least 66 characters,",
			" not counting the terminating newline.\n", Line, "");
	}

/* Merge it into Line: */
for (i = TempFactorStart; i <= TempFactorEnd; i++)
	{
	j = i - TempFactorStart;
	*(Line + i) = *(InsertionString + j);
	}

return 1;
}

/*===========================================================================*/


