/*=============================================================================

				saar.c

			select amino acid range(s)

Select the given range (or set of ranges) of amino acids from a pdb file.

Revision history:
(1)		created 1.3.1998

========includes:============================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "saardefs.h"

/*======defines:=============================================================*/

#define MAXRANGES 20
#define MINLINELENGTH 2

/*======function prototypes:=================================================*/

void		PrintUsage (void);
void		ErrorMessage (char *, char *, char *,
			     char *, char *, char *, char *);
void		WarningMessage (char *, char *, char *,
				char *, char *, char *, char *);
int		LoadRanges (RangeS *, int, char **);
int		ExtractRanges (RangeS *, int, char **);
int		ReadRanges (RangeS *, char *);
int		ParseLine (RangeS *, int, char *);
int		CheckToken (char *);
int		ParseToken (int *, int *, int *, char *);
int		PrepareFileNames (char *, int, char *, int, int, char **);
int		CopyFileNames (char *, int, char *, int, int, char **);
int		CopySelectedRanges (RangeS *, int, char *, char *);
int		RenameFile (char *, char *, int);
int		ProcessLine (RangeS *, int, char *, FILE *);
int		ExtractAminoAcidN (char *);
int		RecogLineType (char *);

/*===========================================================================*/

int main (int argc, char *argv[])
{
RangeS		RangeSA[MAXRANGES + 2];
int		RangesN;
int		RangeFileUsedF;		
char		InFileName[1000], OutFileName[1000];
int		m, n1, n2;
int		AminoAcidsN;

/* At least four arguments are required: */
if (argc < 4)
	{
	PrintUsage ();
	exit (0);
	}

/* If option -f used, read ranges from the specified file: */
RangesN = LoadRanges (RangeSA, argc, argv);

/* If option -f not used, retrieve the ranges from command line arguments: */
if (RangesN == 0)
	{
	RangesN = ExtractRanges (RangeSA, argc, argv);
	RangeFileUsedF = 0;
	}
else RangeFileUsedF = 1;

/* At least one range must be available: */
if (RangesN < 1) exit (0);

/* Prepare file names: */
n1 = sizeof (InFileName) - 1;
n2 = sizeof (OutFileName) - 1;
if (RangeFileUsedF == 1)
	{
	m = PrepareFileNames (InFileName, n1, OutFileName, n2, argc, argv);
	}
else m = CopyFileNames (InFileName, n1, OutFileName, n2, argc, argv);
if (m == -1) exit (0);

/* Finaly, do something useful: */
AminoAcidsN = CopySelectedRanges (RangeSA, RangesN, InFileName, OutFileName);

/* The end: */
return 1;
}

/*======print usage info:====================================================*/

void PrintUsage (void)
{
int		n;

fprintf (stdout, "\n=====================================================\n");
fprintf (stdout, "Select one or more  amino acid  ranges from pdb file.\n");
fprintf (stdout, "=====================================================\n\n");
fprintf (stdout, "Usage:");
fprintf (stdout, "      saar n1:n2 [[-]n3:n4 ...] infile outfile\n");
fprintf (stdout, "            saar -f rangefile infile outfile\n");
fprintf (stdout, "\nn1 is the first amino acid in range,  n2 is the last.\n");
fprintf (stdout, "If option  -f is used, the ranges  are listed  in the\n");
fprintf (stdout, "file named rangefile.  At least one  and no more than\n");
n = MAXRANGES;
fprintf (stdout, "%2d ranges should be specified.\n", n);
fprintf (stdout, "        The minus sign  [-]  means that a given range\n");
fprintf (stdout, "must be excluded.  Do not use  the minus sign  in the\n");
fprintf (stdout, "first range specification.\n");
fprintf (stdout, "        If using  -f option separate ranges by space,\n");
fprintf (stdout, "tab or newline.\n");
fprintf (stdout, "        Residues which have letters  in residue names\n");
fprintf (stdout, "will be treated as if they have no letters,  i.e. the\n");
fprintf (stdout, "letters will be ignored.  Be sure  not to use letters\n");
fprintf (stdout, "when specifying range(s) - only digits  are accepted.\n");
}

/*======print error message:=================================================*/

/* Note: there is no  newline  in the last  four fprintf! */
/* You are expected to supply them through the arguments! */

void ErrorMessage (char *module, char *function, char *filename,
		   char *s1, char *s2, char *s3, char *s4)
{
fprintf (stderr, "\nERROR>");
fprintf (stderr, " module: %s, function: %s", module, function);
if (strlen (filename) != 0) fprintf (stderr, ", file: %s", filename);
fprintf (stderr, "\n");
if (strlen (s1) != 0) fprintf (stderr, "%s", s1);
if (strlen (s2) != 0) fprintf (stderr, "%s", s2);
if (strlen (s3) != 0) fprintf (stderr, "%s", s3);
if (strlen (s4) != 0) fprintf (stderr, "%s", s4);
}

/*======print warning message:===============================================*/

/* Note: there is no  newline  in the last  four fprintf! */
/* You are expected to supply them through the arguments! */

void WarningMessage (char *module, char *function, char *filename,
		     char *s1, char *s2, char *s3, char *s4)
{
fprintf (stdout, "WARNING>");
fprintf (stdout, " module: %s, function: %s", module, function);
if (strlen (filename) != 0) fprintf (stdout, ", file: %s", filename);
fprintf (stdout, "\n");
if (strlen (s1) != 0) fprintf (stdout, "%s", s1);
if (strlen (s2) != 0) fprintf (stdout, "%s", s2);
if (strlen (s3) != 0) fprintf (stdout, "%s", s3);
if (strlen (s4) != 0) fprintf (stdout, "%s", s4);
}

/*======load amino acid range(s) from file:==================================*/

int LoadRanges (RangeS *RSP, int argc, char **argv)
{
int		i, FileNamePosition = -1, RangesN;

/* Check the last argument for -f flag: */
if (strstr (argv[argc - 1], "-f") == argv[argc - 1])
	{
	ErrorMessage ("saar", "LoadRanges", "",
		"File name missing!\n",
		"You tryed to use -f option but failed to\n",
		"specify the name of the file with amino acid ranges.\n", "");
	return -1;
	}

/* Check other arguments for -f flag: */
for (i = 1; i < argc - 1; i++)
	{
	if (strstr (argv[i], "-f") == argv[i]) FileNamePosition = i + 1;
	}

/* If flag -f not found, return 0: */
if (FileNamePosition == -1) return 0;

/* Flag -f was found, check the number of command line arguments: */
if (argc != 5)
	{
	ErrorMessage ("saar", "LoadRanges", "",
		"If using -f option, you have to specify",
		" three files: input pdb file,\noutput pdb",
		" file and file with amino acid ranges",
		" (rangefile)\n");
	return -1;
	}

/* Read ranges from file: */
RangesN = ReadRanges (RSP, argv[FileNamePosition]);
if (RangesN == -1) return -1;

else if (RangesN == 0)
	{
	WarningMessage ("saar", "main", "",
			"Strange, the file with amino acid ranges",
			" specified, but no amino acid numbers\n",
			"found. Check the file format",
			" and contents!\n");
        }
else
	{
	fprintf (stdout, "%d ranges specified. Here they are:\n", RangesN);
	for (i = 0; i < RangesN; i++)
		{
		if (RSP[i].RangeFlag == 1) fprintf (stdout,
				"[%d:%d] included\n",
				 RSP[i].RangeStart, RSP[i].RangeEnd);
		else fprintf (stdout, "[%d:%d] excluded\n",
				 RSP[i].RangeStart, RSP[i].RangeEnd);
		}
	}

return RangesN;
}

/*======read amino acid range(s) from file:==================================*/

int ReadRanges (RangeS *RSP, char *FileName)
{
FILE		*InFile;
char		Line[1000];
int		LineLength;
int		RangesInLine, RangesN = 0;

/* Open file: */
if ((InFile = fopen (FileName, "r")) == NULL)
	{
	ErrorMessage ("saar", "ReadRanges", FileName,
		"Failed to open file!\n",
		"You used -f option; be sure to specify the name\n",
		"of the file with amino acid ranges correctly.\n", "");
	return -1;
	}

/* Read line by line: */
LineLength = sizeof (Line);
while (fgets (Line, LineLength, InFile) != NULL)
	{
	/** Skip comments (line beginning with #): **/
	if (Line[0] == '#') continue;
	/** Parse line: **/
	RangesInLine = ParseLine (RSP, RangesN, Line);
	if (RangesInLine == -1) continue;
	else RangesN += RangesInLine;
	if (RangesN >= MAXRANGES)
		{
		sprintf (Line, "%d", (int) MAXRANGES);
		WarningMessage ("saar", "ReadRanges", FileName,
				"To many ranges specified. ", Line,
				" ranges will be used,",
				" the rest will be ignored.\n");
		break;
		}
	}

/* Close file: */
fclose (InFile);

/* Return the number of amino acid ranges: */
return RangesN;
}

/*======parse line with ranges:==============================================*/

int ParseLine (RangeS *RSP, int RangesN, char *s)
{
char		*p, *tp;
int		m, n1, n2, flag, k, RangesInLine = 0;

/* The initial pointer value (pointer is later set to NULL): */
p = s;

/* Parse line, token by token: */
while ((tp = strtok (p, " \t,;\n")) != NULL)
	{
	/** Ensure that argument to strtok is NULL: **/
	p = NULL;
	/** Check token: **/
	if (CheckToken (tp) == -1) continue;
	/** Parse token (should contain two numbers): **/
	m = ParseToken (&n1, &n2, &flag, tp);
	if (m == -1) continue;
	else
		{
		k = RangesN + RangesInLine;
		/*** Be sure not to exceed array bounds: ***/
		if (k > MAXRANGES - 1) break;
		/*** Set range limits and flag: ***/
		RSP[k].RangeStart = n1;
		RSP[k].RangeEnd = n2;
		RSP[k].RangeFlag = flag;
		/*** Increment index: ***/
		RangesInLine++;
		}
	}

return RangesInLine;
}

/*======in a good token, only digits, minus, and colon are allowed:==========*/

int CheckToken (char *tp)
{
int		c;

while ((c = *tp++) != '\0')
	if ((c != '-') && (c != '+') && (c != ':') && (isdigit (c) == 0))
		return -1;

return 1;
}

/*======parse token (read two positive integers):============================*/

int ParseToken (int *np1, int *np2, int *fp, char *tp)
{
char		*stp;

/* Check is the first character minus sign: */
if (*tp == '-')
	{
	*fp = 0;
	*tp = ' ';
	}
else *fp = 1;

/* Check is the first character + sign; remove it, if present: */
if (*tp == '+') *tp = ' ';

/* Parse token in two subtokens: */
/** The first subtoken: **/
stp = strtok (tp, ":");
if (sscanf (stp, "%d", np1) == EOF) return -1;
if (*np1 < 0) return -1;
/** The second subtoken: **/
stp = strtok (NULL, "\0");
if (sscanf (stp, "%d", np2) == EOF) return -1;
if (*np2 < 0) return -1;
/** Compare range start and end: **/
if (*np1 > *np2) return -1;

return 1;
}

/*======extract amino acid range(s) from command line arguments:=============*/

int ExtractRanges (RangeS *RSP, int argc, char *argv[])
{
int		i, m, n1, n2, flag;
int		RangesN = 0;
char		Line[1000];

/* Scan command line arguments: */
for (i = 1; i < argc - 2; i++)
	{
	/** Check token: **/
	if (CheckToken (argv[i]) == -1) continue;
	/** Parse token (should contain two numbers): **/
	m = ParseToken (&n1, &n2, &flag, argv[i]);
	if (m == -1) continue;
	else
		{
		RSP[RangesN].RangeStart = n1;
		RSP[RangesN].RangeEnd = n2;
		RSP[RangesN].RangeFlag = flag;
		/*** Increment counter: ***/
		RangesN++;
		/*** Be sure not to exceed array bounds: ***/
		if (RangesN >= MAXRANGES)
			{
			sprintf (Line, "%d", (int) MAXRANGES);
			WarningMessage ("saar", "ExtractRanges", "",
					"Too many ranges specified. ",
					Line, " ranges will be used,",
					" the rest will be ignored\n");
			break;
			}
		}
	}

for (i = 0; i < RangesN; i++)
	{
	if (RSP[i].RangeFlag == 1) fprintf (stdout, "[%d:%d] included\n",
			RSP[i].RangeStart, RSP[i].RangeEnd);
	else fprintf (stdout, "[%d:%d] excluded\n",
			RSP[i].RangeStart, RSP[i].RangeEnd);
	}

return RangesN;
}

/*======prepare file names (input and output pdb file):======================*/

int PrepareFileNames (char *InFileName,  int IFNMaxChars,
		      char *OutFileName, int OFNMaxChars,
		      int argc, char **argv)

{
int		i, j, MaxChars;
char		FileName[2][1000];

/* Copy arguments; omit -f flag and the argument which follows that flag: */
j = 0;
i = 1;
MaxChars = sizeof (FileName[0]) - 1;
while (i < argc)
	{
	if (strstr (argv[i], "-f") == argv[i]) i += 2;
	else
		{
		strncpy (FileName[j], argv[i], MaxChars);
		FileName[j][MaxChars] = '\0';
		j++;
		if (j == 2) break;
		i++;
		}
	}

/* Check the number of copied arguments: */
if (j < 2)
	{
	ErrorMessage ("saar", "PrepareFileNames", "",
		"Missing command line arguments.\n",
		"If using -f flag, you have to specify three files.\n",
		"", "");
		return -1;
	}

/* Copy the input and output file name: */
strncpy (InFileName,  FileName[0], IFNMaxChars);
InFileName[IFNMaxChars] = '\0';
strncpy (OutFileName, FileName[1], OFNMaxChars);
OutFileName[OFNMaxChars] = '\0';

/* Compare file names (should be different): */
if (strcmp (InFileName, OutFileName) == 0)
	{
        ErrorMessage ("saar", "PrepareFileNames", InFileName,
                "The same file name is specified",
                " for both input and output file.\n", "", "");
	return -1;
	}

/* Print file names: */
fprintf (stdout, "Input file: %s, output file: %s\n", InFileName, OutFileName);

return 1;
}

/*======copy file names (the last two command line arguments):===============*/

int CopyFileNames (char *InFileName,  int IFNMaxChars,
		   char *OutFileName, int OFNMaxChars,
		   int argc, char **argv)
{
/* Copy the input and output file name: */
strncpy (InFileName,  argv[argc - 2], IFNMaxChars);
strncpy (OutFileName, argv[argc - 1], OFNMaxChars);

/* Compare file names (should be different): */
if (strcmp (InFileName, OutFileName) == 0)
	{
	ErrorMessage ("saar", "CopyFileNames", InFileName,
		"The same file name is specified",
		" for both input and output file.\n", "", "");
	return -1;
	}

/* Print file names: */
fprintf (stdout, "Input file: %s, output file: %s\n", InFileName, OutFileName);

return 1;
}

/*======copy selected ranges:================================================*/

int CopySelectedRanges (RangeS *RSP, int RangesN,
			char *InFileName, char *OutFileName)
{
FILE		*InFile = NULL, *OutFile = NULL;
char		BackupFileName[1000];
int             MaxChars;
int		n;
char            Line[1000];

/* Check the input file: */
if ((InFile = fopen (InFileName, "r")) == NULL)
	{
	ErrorMessage ("saar", "CopySelectedRanges", InFileName,
		"Failed to open the input file.",
		" Check the specified pdb file!\n", "", "");
	return -1;
	}

/* Check the output file (exists allready?): */
if ((OutFile = fopen (OutFileName, "r")) != NULL)
	{
	fclose (OutFile);
	OutFile = NULL;
	MaxChars = (int) sizeof (BackupFileName);
	n = RenameFile (BackupFileName, OutFileName, MaxChars);
	if (n > 0) WarningMessage ("saar", "CopySelectedRanges", OutFileName,
			"The specified output file exists. It will be",
			" renamed to ", BackupFileName, "\n");
	}

/* Line length: */
MaxChars = sizeof (Line);

if ((OutFile = fopen (OutFileName, "w")) == NULL)
	{
	ErrorMessage ("saar", "CopySelectedRanges", OutFileName,
		"Failed to open the output file.",
		" Check who is the owner of this file;\n",
		"check directory, protection",
		" and available disk space.\n");
	if (InFile) fclose (InFile);
	return -1;
	}

/* Read the input file, line by line: */
while (fgets (Line, MaxChars, InFile) != NULL)
	{
	/* Process this line: */
	ProcessLine (RSP, RangesN, Line, OutFile);
	}

/* Close files: */
if (InFile) fclose (InFile);
if (OutFile) fclose (OutFile);

return 1;
}

/*======rename file (put # before the old file name):========================*/

int RenameFile (char *BackupName, char *OldName, int FileNameLen)
{
int		SpaceLeft;

strcpy (BackupName, "#");
SpaceLeft = FileNameLen - strlen (BackupName) - 1;
if (SpaceLeft <= 0) return -1;
strncat (BackupName, OldName, SpaceLeft);
if (rename (OldName, BackupName) != 0) return -1;

return 1;
}

/*======process the line from pdb file:======================================*/

int ProcessLine (RangeS *RSP, int RangesN, char *Line, FILE *OutFile)
{
int		n, AminoAcidN, i, j, IncludeF, RangeF;

/* Recognize the type of line: */
n = RecogLineType (Line);

/* Extract the amino acid number from the line; remove letters, if found: */
if ((AminoAcidN = ExtractAminoAcidN (Line)) == -1) return 0;

/* If line does not contain coordinates write it to the output file; if */
/* it does, check should it be included and written to the output file: */
if (n == 0)
	{
	/** Remove short lines: **/
	if (strlen (Line) > MINLINELENGTH) fputs (Line, OutFile);
	}
else 
	{
	/** Initialize inclusion flag: **/
	IncludeF = 0;
	/** Scan all ranges; if there are contradictory inclusion flags, **/
	/** the last one will be the one which  decides about inclusion: **/
	for (i = 0; i < RangesN; i++)
		{
		RangeF = RSP[i].RangeFlag;
		for (j = RSP[i].RangeStart; j <= RSP[i].RangeEnd; j++)
			{
			if (AminoAcidN == j)
				{
				if (RangeF == 1) IncludeF = 1;
				else IncludeF = 0;
				}
			} 
		}
	/** If inclusion flag is set to 1, write line to the output file: **/
	if (IncludeF == 1) fputs (Line, OutFile);
	}

return 1;
}

/*======recognize line type (coordinates or something else):=================*/

int RecogLineType (char *Line)
{
char            AtomID[] = "ATOM";

/* Does this line refer to atomic coordinates? */
if (strstr (Line, AtomID) == Line) return 1;

return 0;
}

/*======extract the amino acid number; remove any letters:===================*/

int ExtractAminoAcidN (char *Line)
{
int		i, NumberOffset = 22, NumberLength = 5;
char		String[10];
int		c;
int		AminoAcidN;

/* Copy the amino acid number substring: */
for (i = 0; i < NumberLength; i++) String[i] = *(Line + NumberOffset + i);
String[NumberLength] = '\0';

/* Remove trailing characters: */
for (i = NumberLength - 1; i > -1; i--)
	{
	c = String[i];
	if (isdigit (c) == 0) String[i] = ' ';
	}

/* Convert string to number: */
if (sscanf (String, "%d", &AminoAcidN) == EOF) return -1;

return AminoAcidN;
}

/*===========================================================================*/


