/*=============================================================================

				scacor.c

		      Scale correlation coefficient

Purpose: Calculate the correlation coefficient for two hydrophobicity scales,
	 specified by scale number. Scales are read from a file.
	 
Revision history:
(1)		27.01.1999	created

==========includes:==========================================================*/

#include <stdio.h>
#include <string.h>
#include <math.h>

/*======defines:=============================================================*/

#define STRINGSIZE 1000

/*======function prototypes:=================================================*/

void		PrintUsage_ (void);
void		PrintHelp_ (void);
void		ErrorMessage_ (char *, char *, char *,
			       char *, char *, char *, char *);
void		WarningMessage_ (char *, char *, char *,
				 char *, char *, char *, char *);
int		ExtractID_ (char *);
int		RawJob_ (int, int, char *);
int		ExtractValues_ (double *, char *);
double		CorrelCoefficient_ (double *, double *);

/*======main:================================================================*/

int main (int argc, char *argv[])
{
int		n, n1, n2;

/* Check is help requested: */
for (n = 1; n < argc; n++)
	{
	if (strstr (argv[n], "-h") == argv[n])
		{
		PrintHelp_ ();
		exit (0);
		}
	}

/* Four command line arguments expected: */
if (argc != 4)
	{
	fprintf (stderr, "\nFour command line arguments expected!\n");
	PrintUsage_ ();
	exit (-1);
	}

/* Extract ID's: */
if ((n1 = ExtractID_ (argv[1])) < 0) exit (-2);
if ((n2 = ExtractID_ (argv[2])) < 0) exit (-3);

/* The real job: */
if (RawJob_ (n1, n2, argv[3]) < 0) exit (-4);

return 1;
}

/*======usage info:==========================================================*/

void PrintUsage_ (void)
{
fprintf (stdout,
    "\n============================================================\n");
fprintf (stdout,
    "Calculate correlation coefficient for two hydrophob. scales.\n");
fprintf (stdout,
    "============================================================\n\n");
fprintf (stdout,
    "Usage:       scacor n1 n2 infile\n");
fprintf (stdout,
    "             scacor -h    ... print help\n\n");
fprintf (stdout,
    "n1       ... The first scale ID (number).\n");
fprintf (stdout,
    "n2       ... The second scale ID.\n");
fprintf (stdout,
    "infile   ... Input file name (hydrophobicity scales).\n\n");
fprintf (stdout,
    "The correlation coefficient is printed to stdout.  The first\n");
fprintf (stdout,
    "token in input file should be the scale ID;  the next twenty\n");
fprintf (stdout,
    "tokens should contain  hydrophobicity  values.  Extra tokens\n");
fprintf (stdout,
    "are ignored. Tokens  should be  separated by space or comma.\n");
}

/*======detailed information:================================================*/

void PrintHelp_ (void)
{
fprintf (stdout,
   "==============================================================\n");
fprintf (stdout,
   "The purpose of  this program is  to calculate  the correlation\n");
fprintf (stdout,
   "coefficient for two hydrophobicity scales. The formula used to\n");
fprintf (stdout,
   "calculate  the correlation coefficient  for scales x and y is:\n\n");
fprintf (stdout,
   "                 sum (x[i] - <x>) * (y[i] - <y>)\n"); 
fprintf (stdout,
   "r = ----------------------------------------------------------\n");
fprintf (stdout,
   "         |~~~~~~~~~~~~~~~~~~~~~~     |~~~~~~~~~~~~~~~~~~~~~~\n");
fprintf (stdout,
   "      \\  | sum (x[i] - <x>) ^ 2   \\  | sum (y[i] - <y>) ^ 2\n");
fprintf (stdout,
   "        \\|                          \\|\n\n");
fprintf (stdout,
   "... i is the summation index, <x> and <y> are  average values.\n");
fprintf (stdout,
   "--------------------------------------------------------------\n");
fprintf (stdout,
   "Start  scacor  without  any  arguments  to obtain  usage info.\n");
fprintf (stdout,
   "==============================================================\n");
}

/*======print error message:=================================================*/

/* Note: there is no  newline  in the last  four fprintf! */
/* You are expected to supply them through the arguments! */

void ErrorMessage_ (char *module, char *function, char *filename,
		    char *s1, char *s2, char *s3, char *s4)
{
fprintf (stderr, "\nERROR>");
fprintf (stderr, " module: %s, function: %s", module, function);
if (strlen (filename) != 0) fprintf (stderr, ", file: %s", filename);
fprintf (stderr, "\n");
if (strlen (s1) != 0) fprintf (stderr, "%s", s1);
if (strlen (s2) != 0) fprintf (stderr, "%s", s2);
if (strlen (s3) != 0) fprintf (stderr, "%s", s3);
if (strlen (s4) != 0) fprintf (stderr, "%s", s4);
}

/*======print warning message:===============================================*/

/* Note: there is no  newline  in the last  four fprintf! */
/* You are expected to supply them through the arguments! */

void WarningMessage_ (char *module, char *function, char *filename,
		      char *s1, char *s2, char *s3, char *s4)
{
fprintf (stdout, "WARNING>");
fprintf (stdout, " module: %s, function: %s", module, function);
if (strlen (filename) != 0) fprintf (stdout, ", file: %s", filename);
fprintf (stdout, "\n");
if (strlen (s1) != 0) fprintf (stdout, "%s", s1);
if (strlen (s2) != 0) fprintf (stdout, "%s", s2);
if (strlen (s3) != 0) fprintf (stdout, "%s", s3);
if (strlen (s4) != 0) fprintf (stdout, "%s", s4);
}

/*======extract number:======================================================*/

int ExtractID_ (char *s)
{
int		n;

if (sscanf (s, "%d", &n) != 1)
	{
	ErrorMessage_ ("scacor", "ExtractID", "",
		"Failed to extract the first scale ID!\n", "", "", "");
	return -1;
	}

if (n < 0)
	{
	ErrorMessage_ ("scacor", "ExtractID", "",
		"The first scale ID is negative!\n", "", "", "");
	return -2;
	}

return n;
}

/*======the real job:========================================================*/

int RawJob_ (int n1, int n2, char *s)
{
FILE		*fileP;
char		lineA[STRINGSIZE], line_copyA[STRINGSIZE];
int		line_size;
char		*p;
int		n;
int		curr_index;
double		valuesA[2][20];
int		scalesN = 0;
double		correl_coeff;

line_size = sizeof (lineA);

/* Open file for reading: */
if ((fileP = fopen (s, "r")) == NULL)
	{
	ErrorMessage_ ("scacor", "RawJob", "",
		"Failed to open input file!\n", "", "", "");
	return -1;
	}

/* Read file, line by line: */
while (fgets (lineA, line_size, fileP))
	{
	/* Skip comments: */
	if (strstr (lineA, "#") == lineA) continue;

	/* Skip empty lines and very short lines: */
	if (strlen (lineA) <= 1) continue;

	/* Copy line: */
	strncpy (line_copyA, lineA, STRINGSIZE - 1);
	line_copyA[STRINGSIZE - 1] = '\0';

	/* Extract and parse the first token: */
	if ((p = strtok (line_copyA, " ,;\t")) == NULL) continue;
	if (sscanf (p, "%d", &n) != 1) continue;

	/* Check is it one of required ID's: */
	if (n == n1) curr_index = 0;
	else if (n == n2) curr_index = 1;
	else continue;

	/* If this point is reached, extract hydrophobicity values: */
	if (ExtractValues_ (valuesA[curr_index], lineA) < 0) continue;

	/* Values are read, increase count: */
	scalesN++;

	/* If both scales are read, break: */
	if (scalesN == 2) break;
	}

fclose (fileP);

/* Scales read? */
if (scalesN != 2)
	{
	ErrorMessage_ ("scacor", "RawJob", "",
		"Failed to extract requested scales!\n", "", "", "");
	return -1;
	}

/* Calculate the correlation coefficient: */
correl_coeff = CorrelCoefficient_ (valuesA[0], valuesA[1]);

/* Print correlation coefficient to stdout: */
printf ("Correlation coefficient for scales #%d and #%d: %8.3f\n",
	 n1, n2, correl_coeff);

return 1;
}

/*======extract hydrophobicity values:=======================================*/

int ExtractValues_ (double *valuesP, char *s)
{
int		i;
char		*p;
double		value;

/* Remove the first token: */
if ((p = strtok (s, " ,;\t")) == NULL) return -1;

/* Read twenty tokens: */
for (i = 0; i < 20; i++)
	{
	/* Take the next token: */
	if ((p = strtok (NULL, " ,;\t")) == NULL) return -2;

	/* Read value: */
	if (sscanf (p, "%lf", &value) != 1) return -3;

	/* Store value: */
	*(valuesP + i) = value;
	}

/* Return positive value on success: */
return 1;
}

/*======calculate the correlation coefficient:===============================*/

double CorrelCoefficient_ (double *xP, double *yP)
{
double		correl_coeff;
int		i;
double		x_average = 0.0, y_average = 0.0;
double		dx, dy, dxdy, dxdx, dydy, a = 0.0, b = 0.0, c = 0.0;

/* Calculate average values: */
for (i = 0; i < 20; i++)
	{
	x_average += *(xP + i);
	y_average += *(yP + i);
	}
x_average /= 20.0;
y_average /= 20.0;

/* Auxilliary values: */
for (i = 0; i < 20; i++)
	{
	dx = *(xP + i) - x_average;
	dy = *(yP + i) - y_average;
	dxdx = dx * dx;
	dydy = dy * dy;
	dxdy = dx * dy;
	a += dxdy;
	b += dxdx;
	c += dydy;
	}
if (b <= 0)
	{
	ErrorMessage_ ("scacor", "CorrelCoefficient_", "",
		"The first scale contains only zeros!\n", "", "", "");
	return -9999.0;
	}
b = sqrt (b);
if (c <= 0)
	{
	ErrorMessage_ ("scacor", "CorrelCoefficient_", "",
		"The second scale contains only zeros!\n", "", "", "");
	return -9999.0;
	}
c = sqrt (c);

/* Calculate the correlation coefficient: */
correl_coeff = a / (b * c);

return correl_coeff;
}

/*===========================================================================*/


