/*=============================================================================

				    seq2num.c

	Read protein sequence and assign numeric values to amino acids.

Purpose: Read sequence and assign numeric value to each  amino acid. Write
	 amino acid  (serial)  number and numeric  value to  output  file.
	 Numeric values are read from map file, if such file is specified.
	 If not, built-in values are used.

Revision history:
(1)		18.11.1998	created
(2)		22.11.1998	the first version completed

==========includes:==========================================================*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "seq2num.h"

/*======defines:=============================================================*/

/* See seq2num.h */

/*======function prototypes:=================================================*/

void		PrintUsage_ (void);
void		PrintHelp_ (void);
void		ErrorMessage_ (char *, char *, char *,
			       char *, char *, char *, char *);
void		WarningMessage_ (char *, char *, char *,
				 char *, char *, char *, char *);
int		ReadMapFile_ (double *, AminoAcidNameS *, char *);
int		ExtractDigits_ (char *);
int		RawJob_ (char *, char *, double *, AminoAcidNameS *);
int		RecogFileFormat_ (char *);
int		IsFASTA_ (char *);
int		IsEMBL_ (char *);
int		IsSWISSPROT_ (char *);
int		IsPIR_ (char *);
int		IsOneLetterCode_ (char *);
int		IsThreeLettersCode_ (char *);
int		IsVerboseFormat_ (char *);
char		*RevertString_ (char *);
int		IsInteger_ (char *);
int		IsAlphabetic_ (char *);
int		ReadFASTA_ (char *, char *, double *, AminoAcidNameS *);
int		ReadEMBL_ (char *, char *, double *, AminoAcidNameS *);
int		ReadOneLetterCode_ (char *,
				    char *, double *, AminoAcidNameS *);
int		ReadThreeLettersCode_ (char *,
				       char *, double *, AminoAcidNameS *);
int		ReadVerboseFile_ (char *,
				  char *, double *, AminoAcidNameS *);
int		ParseOLCLine_ (FILE *, int *,
			       double *, AminoAcidNameS *, char *);
int		ParseTLCLine_ (FILE *, int *,
			       double *, AminoAcidNameS *, char *);
int		ParseVerboseLine_ (FILE *, int *,
				   double *, AminoAcidNameS *, char *);
int		ReadRawFile_ (char *,
			  char *, double *, AminoAcidNameS *);

/*======main:================================================================*/

int main (int argc, char *argv[])
{
int		n;
char		InFileName[STRINGSIZE], OutFileName[STRINGSIZE],
		MapFileName[STRINGSIZE];
int		size;
AminoAcidNameS	namesSA[23] = {
		{"A", "ALA"}, {"C", "CYS"}, {"L", "LEU"}, {"M", "MET"},
		{"E", "GLU"}, {"Q", "GLN"}, {"H", "HIS"}, {"K", "LYS"},
		{"V", "VAL"}, {"I", "ILE"}, {"F", "PHE"}, {"Y", "TYR"},
		{"W", "TRP"}, {"T", "THR"}, {"G", "GLY"}, {"S", "SER"},
		{"D", "ASP"}, {"N", "ASN"}, {"P", "PRO"}, {"R", "ARG"},
		{"B", "ASX"}, {"Z", "GLX"}, {"X", "UNK"}};
double		valuesA[23] = {
		 1.8,  2.5,  3.8,  1.9,
		-3.5, -3.5, -3.2, -3.9,
		 4.2,  4.5,  2.8, -1.3,
		-0.9, -0.7, -0.4, -0.8,
		-3.5, -3.5, -1.6, -4.5,
		-3.5, -3.5,  0.0};

size = STRINGSIZE;

/* Check is help requested: */
for (n = 1; n < argc; n++)
	{
	if (strstr (argv[n], "-h") == argv[n])
		{
		PrintHelp_ ();
		exit (0);
		}
	}

/* At least three command line arguments expected: */
if (argc < 3)
	{
	fprintf (stderr,
		"\nAt least three command line arguments expected!\n");
	PrintUsage_ ();
	exit (-1);
	}

/* Warn user if too many arguments specified: */
if (argc > 4)
	{
	printf ("\nNo more than 4 command line arguments expected,");
	printf (" but %d found.\n", argc);
	n = argc - 4;
	argc = 4;
	if (n == 1) printf ("The last argument ignored ...\n");
	else printf ("The last %d arguments ignored ...\n", n);
	}

/* Copy the file names: */
strncpy (InFileName, argv[1], size - 1);
InFileName[size - 1] = '\0';
strncpy (OutFileName, argv[2], size - 1);
OutFileName[size - 1] = '\0';
if (argc == 4)
	{
	strncpy (MapFileName, argv[3], size - 1);
	MapFileName[size - 1] = '\0';
	}

/* Read map file and replace built-in numeric values, if map file specified: */
if (argc >= 4)
	{
	n = ReadMapFile_ (valuesA, namesSA, MapFileName);
	if (n < 0) exit (-1);
	}

/* Finally, do the raw job: */
if (RawJob_ (OutFileName, InFileName, valuesA, namesSA) < 0) exit (-2);

/* The end: */
return 1;
}

/*======usage info:==========================================================*/

void PrintUsage_ (void)
{
fprintf (stdout, "\n=================================================\n");
fprintf (stdout, "Assign selected numeric data to a given sequence.\n");
fprintf (stdout, "=================================================\n\n");
fprintf (stdout, "Usage:             seq2num infile outfile\n");
fprintf (stdout, "                   seq2num infile outfile mapfile\n");
fprintf (stdout, "                   seq2num -h      ... print help\n");
}

/*======detailed information:================================================*/

void PrintHelp_ (void)
{
fprintf (stdout,
   "================================================================\n");
fprintf (stdout,
   "The purpose of this program is to prepare the file with selected\n");
fprintf (stdout,
   "numeric data for a  given sequence.  Only one  data item may  be\n");
fprintf (stdout,
   "assigned to each amino acid.  The output file will contain amino\n");
fprintf (stdout,
   "acid number, the associated numeric value,  three  letters  code\n");
fprintf (stdout,
   "and one letter code for each  amino acid.  The output  format is\n");
fprintf (stdout,
   "suitable for  many plotting  programs, especially  for  gnuplot.\n");
fprintf (stdout,
   "Input formats: FASTA, EMBL, SWISSPROT, PIR and raw sequence plus\n");
fprintf (stdout,
   "three formats from exseq program: one letter code, three l. code\n");
fprintf (stdout,
   "and verbose. File is treated as raw if format is not recognized.\n");
fprintf (stdout,
   "----------------------------------------------------------------\n");
fprintf (stdout, "Usage:             seq2num infile outfile\n");
fprintf (stdout, "                   seq2num infile outfile mapfile\n");
fprintf (stdout, "                   seq2num -h      ... print help\n");
fprintf (stdout,
   "mapfile  is the file with numeric values. If not  specified, the\n");
fprintf (stdout,
   "built-in values  based on  Kyte-Doolittle  hydrophobicity  scale\n");
fprintf (stdout,
   "will be used.  The mapfile should  contain  amino acid  names in\n");
fprintf (stdout,
   "three letters code,  accompained  with numeric values.  Example:\n");
fprintf (stdout, "ARG 70.0\nLYS 60.0\n");
fprintf (stdout,
   "Zero value  is assigned to amino acids  not listed in  map file.\n");
fprintf (stdout,
   "================================================================\n");
}

/*======print error message:=================================================*/

/* Note: there is no  newline  in the last  four fprintf! */
/* You are expected to supply them through the arguments! */

void ErrorMessage_ (char *module, char *function, char *filename,
		    char *s1, char *s2, char *s3, char *s4)
{
fprintf (stderr, "\nERROR>");
fprintf (stderr, " module: %s, function: %s", module, function);
if (strlen (filename) != 0) fprintf (stderr, ", file: %s", filename);
fprintf (stderr, "\n");
if (strlen (s1) != 0) fprintf (stderr, "%s", s1);
if (strlen (s2) != 0) fprintf (stderr, "%s", s2);
if (strlen (s3) != 0) fprintf (stderr, "%s", s3);
if (strlen (s4) != 0) fprintf (stderr, "%s", s4);
}

/*======print warning message:===============================================*/

/* Note: there is no  newline  in the last  four fprintf! */
/* You are expected to supply them through the arguments! */

void WarningMessage_ (char *module, char *function, char *filename,
		      char *s1, char *s2, char *s3, char *s4)
{
fprintf (stdout, "WARNING>");
fprintf (stdout, " module: %s, function: %s", module, function);
if (strlen (filename) != 0) fprintf (stdout, ", file: %s", filename);
fprintf (stdout, "\n");
if (strlen (s1) != 0) fprintf (stdout, "%s", s1);
if (strlen (s2) != 0) fprintf (stdout, "%s", s2);
if (strlen (s3) != 0) fprintf (stdout, "%s", s3);
if (strlen (s4) != 0) fprintf (stdout, "%s", s4);
}

/*======read map file:=======================================================*/

int ReadMapFile_ (double *valuesP, AminoAcidNameS *namesSP, char *FileName)
{
FILE		*File;
char		Line[STRINGSIZE];
int		size;
int		n;
double		value;

if ((File = fopen (FileName, "r")) == NULL)
	{
	ErrorMessage_ ("seq2num", "ReadMapFile", FileName,
		"Failed to open map file!\n",
		"Check the file name and path.\n",
		"", "");
	return -1;
	}

/* Reset values: */
for (n = 0; n < 23; n++) *(valuesP + n) = 0.0;

/* Read line by line: */
size = sizeof (Line);
while (fgets (Line, size, File) != NULL)
	{
	/** Skip comments (line beginning with #): **/
	if (Line[0] == '#') continue;

	/** Remove newline, if present: **/
	n = strlen (Line);
	if (Line[n - 1] == '\n') Line[n - 1] = '\0';

	/** Skip empty lines: **/
	if (strlen (Line) == 0) continue;

	/** Scan all amino acid names: **/
	for (n = 0; n < 23; n++)
		{
		if (!strstr (Line, (namesSP + n)->LongName)) continue;
		ExtractDigits_ (Line);
		if (sscanf (Line, "%lf", &value) != 1) continue;
		*(valuesP + n) = value;
		break;
		}
	}

fclose (File);
return 1;
}

/*======extract digits, minus signs and decimal points:======================*/

int ExtractDigits_ (char *s)
{
int		count = 0;
char		*p;
int		n;

p = s;
while ((n = *p++) != '\0')
	{
	if (isdigit (n)) continue;
	if (n == '-') continue;
	if (n == '.') continue;
	*(p - 1) = ' ';
	count++;
	}

/* Return the number of characters replaced: */
return count;
}

/*======raw job (read sequence, map values, write output file):==============*/

int RawJob_ (char *OutFileName, char *InFileName,
	     double *valuesP, AminoAcidNameS *NSP)
{
int		aaN;
int		n;

/* Try to recognize the file format: */
if ((n = RecogFileFormat_ (InFileName)) < 0)
	{
	WarningMessage_ ("seq2num", "RawJob", InFileName,
		"Unable to recognize input file format -",
		" raw file (one letter code) assumed!\n", "", "");
	}

/* Inform the user about input file format: */
if      (n == 0) printf ("Input file contains data in FASTA format ...\n");
else if (n == 1) printf ("Input file contains data in EMBL format ...\n");
else if (n == 2) printf ("Input file contains data in SWISSPROT format ...\n");
else if (n == 3) printf ("Input file contains data in PIR format ...\n");
else if (n == 4) printf ("Input file contains one letter code ...\n");
else if (n == 5) printf ("Input file contains three letters code ...\n");
else if (n == 6) printf ("Input file contains verbose information ...\n");

/* Read input file, write output file: */
/** Important note: the functions used below are very similar. **/
/** For example, they open files in the same way, read line by **/
/** line etc. The reason files are not opened in this (RawJob) **/
/** function is that  some files  may  be  binary,  so  future **/
/** extensions will be complicated if file open operations are **/
/** performed  with  assumption that  file  format is  "text". **/
switch (n)
	{
	case 0:
		aaN = ReadFASTA_ (OutFileName, InFileName, valuesP, NSP);
		break;
	case 1:
	case 2:
	case 3:
		aaN = ReadEMBL_ (OutFileName, InFileName, valuesP, NSP);
		break;
	case 4:
		aaN = ReadOneLetterCode_ (OutFileName,
					  InFileName, valuesP, NSP);
		break;
	case 5:
		aaN = ReadThreeLettersCode_ (OutFileName,
					     InFileName, valuesP, NSP);
		break;
	case 6:
		aaN = ReadVerboseFile_ (OutFileName, InFileName, valuesP, NSP);
		break;
	default:
		aaN = ReadRawFile_ (OutFileName, InFileName, valuesP, NSP);
	}

/* Return the number of amino acids: */
return aaN;
}

/*======recognize input file format:=========================================*/
/* -1 = format not recognized */
/*  0 = one letter code, 40 codes per line */
/*  1 = three letters code, 20 codes per line */
/*  2 = verbose three letters format */
/* for further details, see the exseq program! */

int RecogFileFormat_ (char *InFileName)
{
int		n;

if ((n = IsFASTA_ (InFileName)) >= 0) return n;
if ((n = IsEMBL_ (InFileName)) >= 0) return n;
if ((n = IsSWISSPROT_ (InFileName)) >= 0) return n;
if ((n = IsPIR_ (InFileName)) >= 0) return n;
if ((n = IsOneLetterCode_ (InFileName)) >= 0) return n;
if ((n = IsThreeLettersCode_ (InFileName)) >= 0) return n;
if ((n = IsVerboseFormat_ (InFileName)) >= 0) return n;

return -1;
}

/*======check is FASTA sequence in input file:===============================*/

int IsFASTA_ (char *FileName)
{
FILE		*File;
char		Line[STRINGSIZE];
int		size;
int		n;

size = sizeof (Line);

if (!(File = fopen (FileName, "r")))
	{
	ErrorMessage_ ("seq2num", "IsOneLetterCode", FileName,
		"Failed to open input file!\n",
		"", "", "");
	return -1;
	}

/* The first line should contain > as the first character: */
if (!fgets (Line, size, File))
        {
        ErrorMessage_ ("seq2num", "IsFASTA", FileName,
                "The input file is empty!\n",
                "", "", "");
        fclose (File);
        return -2;
        }
if (strstr (Line, ">") != Line)
	{
	fclose (File);
	return -3;
	}

/* Check is there anything left at all: */
if (!fgets (Line, size, File))
	{
	fclose (File);
	return -4;
	}

/* Close file: */
fclose (File);

/* Remove trailing newline: */
n = strlen (Line);
if (Line[n - 1] == '\n') Line[n - 1] = '\0';

/* The second line should contain only alphabetic characters: */
if (!IsAlphabetic_ (Line)) return -5;

/* If this point is reached, FASTA file is in question: */
return 0;
}

/*======check is EMBL file in question:======================================*/

int IsEMBL_ (char *FileName)
{
FILE		*File;
char		Line[STRINGSIZE];
int		size;
int		n;
int		ItemFoundF = 0;
char		*TokenP;

size = sizeof (Line);

if (!(File = fopen (FileName, "r")))
	{
	ErrorMessage_ ("seq2num", "IsEMBL", FileName,
		"Failed to open input file!\n",
		"", "", "");
	return -1;
	}

/* Read everything up to SQ: */
while (fgets (Line, size, File))
	{
	if (strstr (Line, "SQ") == Line)
		{
		ItemFoundF = 1;
		break;
		}
	}

/* If SQ not found, it is not EMBL format: */
if (!ItemFoundF)
	{
	fclose (File);
	return -2;
	}

/* The line after SQ should have digits in the last token: */
if (!fgets (Line, size, File))
	{
	fclose (File);
	return -3;
	}
n = strlen (Line);
if (Line[n - 1] == '\n') Line[n - 1] = '\0';
RevertString_ (Line);
if ((TokenP = strtok (Line, " ")) == NULL)
	{
	fclose (File);
	return -4;
	}
if (!IsInteger_ (TokenP))
	{
	fclose (File);
	return -5;
	}

/* Read everything up to //: */
ItemFoundF = 0;
while (fgets (Line, size, File))
	{
	if (strstr (Line, "//") == Line)
		{
		ItemFoundF = 1;
		break;
		}
	}

/* Close file: */
fclose (File);

/* Return negative on failure: */
if (ItemFoundF == 0) return -6;

/* If this point is reached, EMBL file is recognized: */
return 1;
}

/*======check is SWISSPROT file in question:=================================*/

int IsSWISSPROT_ (char *FileName)
{
FILE		*File;
char		Line[STRINGSIZE];
int		size;
int		n;
int		ItemFoundF = 0;
char		*TokenP;

size = sizeof (Line);

if (!(File = fopen (FileName, "r")))
	{
	ErrorMessage_ ("seq2num", "IsSWISSPROT", FileName,
		"Failed to open input file!\n",
		"", "", "");
	return -1;
	}

/* Read everything up to SQ: */
while (fgets (Line, size, File))
	{
	if (strstr (Line, "SQ") == Line)
		{
		ItemFoundF = 1;
		break;
		}
	}

/* If SQ not found, it is not SWISSPROT format: */
if (!ItemFoundF)
	{
	fclose (File);
	return -2;
	}

/* The line after SQ should have alphabetic characters in the last token: */
if (!fgets (Line, size, File))
	{
	fclose (File);
	return -3;
	}
n = strlen (Line);
if (Line[n - 1] == '\n') Line[n - 1] = '\0';
RevertString_ (Line);
if ((TokenP = strtok (Line, " ")) == NULL)
	{
	fclose (File);
	return -4;
	}
if (!IsAlphabetic_ (TokenP))
	{
	fclose (File);
	return -5;
	}

/* Read everything up to //: */
ItemFoundF = 0;
while (fgets (Line, size, File))
	{
	if (strstr (Line, "//") == Line)
		{
		ItemFoundF = 1;
		break;
		}
	}

/* Close file: */
fclose (File);

/* Return negative on failure: */
if (ItemFoundF == 0) return -6;

/* If this point is reached, SWISSPROT file is recognized: */
return 2;
}

/*======check is PIR file in question:=======================================*/

int IsPIR_ (char *FileName)
{
FILE		*File;
char		Line[STRINGSIZE];
int		size;
int		n;
int		ItemFoundF = 0;
char		*TokenP;

size = sizeof (Line);

if (!(File = fopen (FileName, "r")))
	{
	ErrorMessage_ ("seq2num", "IsPIR", FileName,
		"Failed to open input file!\n",
		"", "", "");
	return -1;
	}

/* Read everything up to SEQUENCE: */
while (fgets (Line, size, File))
	{
	if (strstr (Line, "SEQUENCE") == Line)
		{
		ItemFoundF = 1;
		break;
		}
	}

/* If SEQUENCE not found, it is not PIR format: */
if (!ItemFoundF)
	{
	fclose (File);
	return -2;
	}

/* The line after SEQUENCE should have digits in the first token: */
if (!fgets (Line, size, File))
	{
	fclose (File);
	return -3;
	}
n = strlen (Line);
if (Line[n - 1] == '\n') Line[n - 1] = '\0';
if ((TokenP = strtok (Line, " ")) == NULL)
	{
	fclose (File);
	return -4;
	}
if (!IsInteger_ (TokenP))
	{
	fclose (File);
	return -5;
	}

/* Read everything up to ///: */
ItemFoundF = 0;
while (fgets (Line, size, File))
	{
	if (strstr (Line, "///") == Line)
		{
		ItemFoundF = 1;
		break;
		}
	}

/* Close file: */
fclose (File);

/* Return negative on failure: */
if (ItemFoundF == 0) return -6;

/* If this point is reached, PIR file is recognized: */
return 3;
}

/*======check is one letter code in input file:==============================*/
/* Valid lines are: */
/* nonempty */
/* containing only alphabetic characters */
/* no more than 40 characters long, not counting newline */
/* containing no spaces or tabs */
/* Return value: 4 on success, negative on failure */

int IsOneLetterCode_ (char *FileName)
{
FILE		*File;
char		Line[STRINGSIZE];
int		size;
int		i, n;
int		GoodLineF;

size = sizeof (Line);

if (!(File = fopen (FileName, "r")))
	{
	ErrorMessage_ ("seq2num", "IsOneLetterCode", FileName,
		"Failed to open input file!\n",
		"", "", "");
	return -1;
	}

/* The first line should contain # as the first character: */
if (!fgets (Line, size, File))
	{
	ErrorMessage_ ("seq2num", "IsOneLetterCode", FileName,
		"The input file is empty!\n",
		"", "", "");
	fclose (File);
	return -2;
	}
if (strstr (Line, "#") != Line) return -3;

/* Check line by line; return 4 if at least one good line is found: */
while (fgets (Line, size, File))
	{
	n = strlen (Line);

	/** Do not check the terminal newline, if present: **/
	if (Line[n - 1] == '\n')
		{
		Line[n - 1] = '\0';
		n--;
		}

	/** Skip titles and comments: **/
	if (Line[0] == '#') continue;

	/** Skip empty lines: **/
	if (n == 0) continue;

	/** Valid lines contain no more than 40 characters: **/
	if (n > 40) break;

	/** Valid lines contain only alphabetic characters but no spaces: **/	
	GoodLineF = 1;
	for (i = 0; i < n; i++)
		{
		if (Line[i] == ' ')
			{
			GoodLineF = 0;
			break;
			}
		if (!isalpha (Line[i]))
			{
			GoodLineF = 0;
			break;
			}
		}
	if (!GoodLineF) continue;

	/** A special case (three letter code alone in line): **/
	if (strlen (Line) == 3) continue;

	/** If this point is reached, good line was found: **/
	fclose (File);
	return 4;
	}

fclose (File);

return -4;
}

/*======check is three letters code in input file:===========================*/
/* Valid lines: */
/* contain no more than 79 characters */
/* contain no more than 20 tokens, separated by space */
/* contain only alphabetic characters */
/* each token consists of three characters */
/* Return value: 5 on success, negative on failure */

int IsThreeLettersCode_ (char *FileName)
{
FILE		*File;
char		Line[STRINGSIZE];
int		size;
int		i, n;
int		GoodLineF;

size = sizeof (Line);

if (!(File = fopen (FileName, "r")))
	{
	ErrorMessage_ ("seq2num", "IsThreeLettersCode", FileName,
		"Failed to open input file!\n",
		"", "", "");
	return -1;
	}

/* Check line by line; return 5 if at least one good line is found: */
while (fgets (Line, size, File))
	{
	n = strlen (Line);

	/** Do not check the terminal newline, if present: **/
	if (Line[n - 1] == '\n')
		{
		Line[n - 1] = '\0';
		n--;
		}

	/** Skip titles and comments: **/
	if (Line[0] == '#') continue;

	/** Skip empty lines: **/
	if (n == 0) continue;

	/** Valid lines contain no more than 79 characters: **/
	if (n > 79) continue;

	/** Check for spaces and non-alphabetic characters: **/
	GoodLineF = 1;
	for (i = 0; i < n; i++)
		{
		/*** Spaces should be present at positions 4, 8, 12 etc.: ***/
		if ((i + 1) % 4 == 0)
			{
			if (Line[i] != ' ')
				{
				GoodLineF = 0;
				break;
				}
			}
		/*** Other characters should be alphabetic: ***/
		else
			{
			if (!isalpha (Line[i]))
				{
				GoodLineF = 0;
				break;
				}
			if (Line[i] == ' ')
				{
				GoodLineF = 0;
				break;
				}
			}
		}
	if (!GoodLineF) continue;

	/** If this point is reached, good line was found: **/
	fclose (File);
	return 5;
	}

fclose (File);

return -1;
}

/*======check is verbose sequence information in input file:=================*/
/* Valid lines consist of: */
/* Amino acid global index (positive integer) */
/* Amino acid local index (positive integer) */
/* Three letters code */
/* One letter code */
/* Amino acid serial number (positive integer) */
/* Chain ID (one alphabetic character, space allowed) */
/* Return value: 6 on success, negative on failure */

int IsVerboseFormat_ (char *FileName)
{
FILE		*File;
char		Line[STRINGSIZE];
int		size;
int		n;
char		*TokenP;

size = sizeof (Line);

if (!(File = fopen (FileName, "r")))
	{
	ErrorMessage_ ("seq2num", "IsThreeLettersCode", FileName,
		"Failed to open input file!\n",
		"", "", "");
	return -1;
	}

/* Check line by line; return 6 if at least one good line is found: */
while (fgets (Line, size, File))
	{
	n = strlen (Line);

	/** Do not check the terminal newline, if present: **/
	if (Line[n - 1] == '\n')
		{
		Line[n - 1] = '\0';
		n--;
		}

	/** Skip titles and comments: **/
	if (Line[0] == '#') continue;

	/** Skip empty lines: **/
	if (n == 0) continue;

	/** Check the first three tokens: **/
	if ((TokenP = strtok (Line, " \t")) == NULL) break;
	if (!IsInteger_ (TokenP)) break;
	if ((TokenP = strtok (NULL, " \t")) == NULL) break;
	if (!IsInteger_ (TokenP)) break;
	if ((TokenP = strtok (NULL, " \t")) == NULL) break;
	if (!IsAlphabetic_ (TokenP)) break;

	/** If this point is reached, good line was found: **/
	fclose (File);
	return 6;
	}

fclose (File);

return -1;
}

/*======revert string:=======================================================*/

char *RevertString_ (char *s)
{
int		i, m, n;
char		c;

n = strlen (s);
m = n / 2;
for (i = 0; i <= m; i++)
	{
	c = *(s + i);
	*(s + i) = *(s + n + i - 1);
	*(s + n + i - 1) = c;
	}
return s;
}

/*======string contains digits only?=========================================*/

int IsInteger_ (char *s)
{
int		n;

while ((n = *s++) != '\0') if (!isdigit (n)) return 0;

/* If this point is reached, all characters are digits: */
return 1;
}

/*======string contains alphabetic characters only?==========================*/

int IsAlphabetic_ (char *s)
{
int		n;

while ((n = *s++) != '\0') if (!isalpha (n)) return 0;

/* If this point is reached, all characters are digits: */
return 1;
}

/*======read FASTA file:=====================================================*/

int ReadFASTA_ (char *OutFileName, char *InFileName,
		double *valuesP, AminoAcidNameS *NSP)
{
int		aaN = 0;
FILE		*InFile, *OutFile;
char		Line[STRINGSIZE];
int		size, n;

size = sizeof (Line);

if ((OutFile = fopen (OutFileName, "w")) == NULL)
	{
	ErrorMessage_ ("seq2num", "ReadFASTA", OutFileName,
		"Failed to open output file!\n",
		"", "", "");
	return -1;
	}

if ((InFile = fopen (InFileName, "r")) == NULL)
	{
	ErrorMessage_ ("seq2num", "ReadFASTA", InFileName,
		"Failed to open input file!\n",
		"", "", "");
	fclose (OutFile);
	return -2;
	}

/* Ignore the first line: */
if (!fgets (Line, size, InFile))
	{
	ErrorMessage_ ("seq2num", "ReadFASTA", InFileName,
		"Input file is empty!\n",
		"", "", "");
		fclose (OutFile);
	fclose (InFile);
	return -3;
	}

/* Read input file, line by line: */
while (fgets (Line, size, InFile))
	{
	/** Ignore comments: **/
	if (strstr (Line, ">") == Line) continue;

	/** Remove trailing newline: **/
	n = strlen (Line);
	if (Line[n - 1] == '\n') Line[n - 1] = '\0';

	/** If this point is reached, parse the line: **/
	ParseOLCLine_ (OutFile, &aaN, valuesP, NSP, Line);
	}

/* Close files, return the number of amino acids: */
fclose (OutFile);
fclose (InFile);
return aaN;
}

/*======read EMBL or SWISSPROT file:=========================================*/

int ReadEMBL_ (char *OutFileName, char *InFileName,
	       double *valuesP, AminoAcidNameS *NSP)
{
int		aaN = 0;
FILE		*InFile, *OutFile;
char		Line[STRINGSIZE];
int		size, n;

size = sizeof (Line);

if ((OutFile = fopen (OutFileName, "w")) == NULL)
	{
	ErrorMessage_ ("seq2num", "ReadEMBL_", OutFileName,
		"Failed to open output file!\n",
		"", "", "");
	return -1;
	}

if ((InFile = fopen (InFileName, "r")) == NULL)
	{
	ErrorMessage_ ("seq2num", "ReadEMBL_", InFileName,
		"Failed to open input file!\n",
		"", "", "");
	fclose (OutFile);
	return -2;
	}

/* Read input file: */
while (fgets (Line, size, InFile))
	{
	/** Take lines with space or number at the beginning as **/
	/** seq. data (common for  EMBL,  SWISSPROT  and  PIR): **/
	if ((Line[0] != ' ') && (!isdigit (Line[0]))) continue;

        /** Remove trailing newline: **/
        n = strlen (Line);
        if (Line[n - 1] == '\n') Line[n - 1] = '\0';

        /** If this point is reached, parse the line: **/
        ParseOLCLine_ (OutFile, &aaN, valuesP, NSP, Line);
	}

/* Close files, return the number of amino acids: */
fclose (OutFile);
fclose (InFile);
return aaN;
}

/*======read one letter code:================================================*/
/* Includes D. J. format = Davor Juretic format */

int ReadOneLetterCode_ (char *OutFileName, char *InFileName,
			double *valuesP, AminoAcidNameS *NSP)
{
int		aaN = 0;
FILE		*InFile, *OutFile;
char		Line[STRINGSIZE];
int		size, n;

size = sizeof (Line);

if ((OutFile = fopen (OutFileName, "w")) == NULL)
	{
	ErrorMessage_ ("seq2num", "ReadOneLetterCode", OutFileName,
		"Failed to open output file!\n",
		"", "", "");
	return -1;
	}

if ((InFile = fopen (InFileName, "r")) == NULL)
	{
	ErrorMessage_ ("seq2num", "ReadOneLetterCode", InFileName,
		"Failed to open input file!\n",
		"", "", "");
	fclose (OutFile);
	return -2;
	}

/* Ignore the first line (this ensures compatibility with D. J. format): */
if (!fgets (Line, size, InFile))
	{
	ErrorMessage_ ("seq2num", "ReadOneLetterCode", InFileName,
		"Input file is empty!\n",
		"", "", "");
	fclose (OutFile);
	fclose (InFile);
	return -3;
	}

/* Read input file, line by line: */
while (fgets (Line, size, InFile))
	{
	/** Ignore comments: **/
	if (strstr (Line, "#") == Line) continue;

	/** Remove trailing newline: **/
	n = strlen (Line);
	if (Line[n - 1] == '\n') Line[n - 1] = '\0';

	/** If line contains the letter U (or u), stop reading: **/
	/** Note: this is for compatibility with  D. J. format! **/
	if (strstr (Line, "U") || strstr (Line, "u")) break;

	/** If this point is reached, parse the line: **/
	ParseOLCLine_ (OutFile, &aaN, valuesP, NSP, Line);
	}

/* Close files, return the number of amino acids: */
fclose (OutFile);
fclose (InFile);
return aaN;
}

/*======parse line with one letter code, write values to output file:========*/

int ParseOLCLine_ (FILE *OutFile, int *aaNP,
		   double *valuesP, AminoAcidNameS *NSP, char *s)
{
int		delta, oldaaN;
char		*p;
int		i, n;
char		Name[2];

oldaaN = *aaNP;
Name[1] = '\0';
p = s;

while ((n = *p++) != '\0')
	{
	n = toupper (n);
	Name[0] = n;
	for (i = 0; i < 23; i++)
		{
		if (strcmp (Name, (NSP + i)->ShortName) != 0) continue;

		/* If this point is reached, amino acid is recognized; */
		/* increase count and write data to output file: */
		(*aaNP)++;
		fprintf (OutFile, "%6d %8.2f    %3s  %1s\n",
			 *aaNP, *(valuesP + i),
			 (NSP + i)->LongName, (NSP + i)->ShortName);
		}
	}

delta = *aaNP - oldaaN;
return delta;
}

/*======read three letters code:=============================================*/

int ReadThreeLettersCode_ (char *OutFileName, char *InFileName,
			   double *valuesP, AminoAcidNameS *NSP)
{
int		aaN = 0;
FILE		*InFile, *OutFile;
char		Line[STRINGSIZE];
int		size, n;

size = sizeof (Line);

if ((OutFile = fopen (OutFileName, "w")) == NULL)
	{
	ErrorMessage_ ("seq2num", "ReadThreeLettersCode", OutFileName,
		"Failed to open output file!\n",
		"", "", "");
	return -1;
	}

if ((InFile = fopen (InFileName, "r")) == NULL)
	{
	ErrorMessage_ ("seq2num", "ReadThreeLettersCode", InFileName,
		"Failed to open input file!\n",
		"", "", "");
	fclose (OutFile);
	return -2;
	}

/* Read input file, line by line: */
while (fgets (Line, size, InFile))
	{
	/** Ignore comments: **/
	if (strstr (Line, "#") == Line) continue;

	/** Remove trailing newline: **/
	n = strlen (Line);
	if (Line[n - 1] == '\n') Line[n - 1] = '\0';

	/** If this point is reached, parse the line: **/
	ParseTLCLine_ (OutFile, &aaN, valuesP, NSP, Line);
	}

/* Close files, return the number of amino acids: */
fclose (OutFile);
fclose (InFile);
return aaN;
}

/*======parse line with one letter code, write values to output file:========*/

int ParseTLCLine_ (FILE *OutFile, int *aaNP,
                   double *valuesP, AminoAcidNameS *NSP, char *s)
{
int		delta, oldaaN;
char		*p, *CurrP;
int		i;

oldaaN = *aaNP;
CurrP = s;

while ((p = strtok (CurrP, " \t")) != NULL) 
	{
	CurrP = NULL;
	/* Convert to uppercase: */
	*p = toupper (*p);
	*(p + 1) = toupper (*(p + 1));
	*(p + 2) = toupper (*(p + 2));

	for (i = 0; i < 23; i++)
		{
		if (strcmp (p, (NSP + i)->LongName) != 0) continue;

		/* If this point is reached, amino acid is recognized; */
		/* increase count and write data to output file: */
		(*aaNP)++;
		fprintf (OutFile, "%6d %8.2f    %3s  %1s\n",
			 *aaNP, *(valuesP + i),
			 (NSP + i)->LongName, (NSP + i)->ShortName);
		}
	}

delta = *aaNP - oldaaN;
return delta;
}

/*======read the file with verbose information:==============================*/

int ReadVerboseFile_ (char *OutFileName, char *InFileName,
		      double *valuesP, AminoAcidNameS *NSP)
{
int		aaN = 0;
FILE		*InFile, *OutFile;
char		Line[STRINGSIZE];
int		size, n;

size = sizeof (Line);

if ((OutFile = fopen (OutFileName, "w")) == NULL)
	{
	ErrorMessage_ ("seq2num", "ReadVerboseFile", OutFileName,
		"Failed to open output file!\n",
		"", "", "");
	return -1;
	}

if ((InFile = fopen (InFileName, "r")) == NULL)
	{
	ErrorMessage_ ("seq2num", "ReadVerboseFile", InFileName,
		"Failed to open input file!\n",
		"", "", "");
	fclose (OutFile);
	return -2;
	}

/* Read input file, line by line: */
while (fgets (Line, size, InFile))
	{
	/** Ignore comments: **/
	if (strstr (Line, "#") == Line) continue;

	/** Remove trailing newline: **/
	n = strlen (Line);
	if (Line[n - 1] == '\n') Line[n - 1] = '\0';

	/** If this point is reached, parse the line: **/
	ParseVerboseLine_ (OutFile, &aaN, valuesP, NSP, Line);
	}

/* Close files, return the number of amino acids: */
fclose (OutFile);
fclose (InFile);
return aaN;
}

/*======parse verbose line and write values to output file:==================*/

int ParseVerboseLine_ (FILE *OutFile, int *aaNP,
		       double *valuesP, AminoAcidNameS *NSP, char *s)
{
int		delta, oldaaN;
char		*p;
int		i;

oldaaN = *aaNP;

/* Skip two tokens: */
if ((p = strtok (s, " \t")) == NULL) return 0;
if ((p = strtok (NULL, " \t")) == NULL) return 0;

/* The third token should be amino acid name: */
if ((p = strtok (NULL, " \t")) == NULL) return 0;

/* Convert to uppercase: */
*p = toupper (*p);
*(p + 1) = toupper (*(p + 1));
*(p + 2) = toupper (*(p + 2));

/* Identify amino acid by name: */
for (i = 0; i < 23; i++)
	{
	if (strcmp (p, (NSP + i)->LongName) != 0) continue;

	/* If this point is reached, amino acid is recognized; */
	/* increase count and write data to output file: */
	(*aaNP)++;
	fprintf (OutFile, "%6d %8.2f    %3s  %1s\n",
	*aaNP, *(valuesP + i),
	(NSP + i)->LongName, (NSP + i)->ShortName);
	}

delta = *aaNP - oldaaN;
return delta;
}

/*======read raw file (plain sequence, no comments):=========================*/

int ReadRawFile_ (char *OutFileName, char *InFileName,
		  double *valuesP, AminoAcidNameS *NSP)
{
int		aaN = 0;
FILE		*InFile, *OutFile;
char		Line[STRINGSIZE];
int		size, n;

size = sizeof (Line);

if ((OutFile = fopen (OutFileName, "w")) == NULL)
	{
	ErrorMessage_ ("seq2num", "ReadRawFile", OutFileName,
		"Failed to open output file!\n",
		"", "", "");
	return -1;
	}

if ((InFile = fopen (InFileName, "r")) == NULL)
	{
	ErrorMessage_ ("seq2num", "ReadRawFile", InFileName,
		"Failed to open input file!\n",
		"", "", "");
	fclose (OutFile);
	return -2;
	}

/* Read input file, line by line: */
while (fgets (Line, size, InFile))
	{
	/** Remove trailing newline: **/
	n = strlen (Line);
	if (Line[n - 1] == '\n') Line[n - 1] = '\0';

	/** If this point is reached, parse the line: **/
	ParseOLCLine_ (OutFile, &aaN, valuesP, NSP, Line);
	}

/* Close files, return the number of amino acids: */
fclose (OutFile);
fclose (InFile);
return aaN;
}

/*===========================================================================*/


